import pytest

import inferpy as inf
from inferpy.contextmanager import randvar_registry
from inferpy.util import tf_graph


@pytest.fixture(params=[
    # default context, build graph
    (True, True),
    # new context, build graph
    (False, True),
    # new context, not build graph
    (False, False)
])
def init_context(request):
    is_default, is_building = request.param
    result = {
        'is_building': is_building,
        'is_default': is_default
    }
    if not is_default:
        with randvar_registry.init(None if is_building else tf_graph.get_empty_graph()):
            yield result
    else:
        yield result


def test_context(init_context):
    assert randvar_registry.is_building_graph() is init_context['is_building']
    assert randvar_registry.is_default() is init_context['is_default']


def test_register_variable(init_context):
    is_default = init_context['is_default']
    name = 'x'

    # the variable does not exist in the context
    assert randvar_registry.get_variable(name) is None
    assert randvar_registry.get_variable_or_parameter(name) is None
    inf.Normal(0, 1, name=name)

    # the variable exists in the context
    # randvar_registry.register_variable(x) has been automatically called
    assert randvar_registry.get_variable(name) is not None
    assert randvar_registry.get_variable_or_parameter(name) is not None

    # if create a new variable with the same name, it fails just if is_default is False
    if is_default:
        inf.Normal(0, 1, name=name)
        assert randvar_registry.get_variable(name) is not None
        assert randvar_registry.get_variable_or_parameter(name) is not None
    else:
        with pytest.raises(ValueError):
            inf.Normal(0, 1, name=name)


def test_register_parameter(init_context):
    is_default = init_context['is_default']
    name = 'x'

    # the parameter does not exist in the context
    assert len(randvar_registry.get_var_parameters()) == 0
    assert randvar_registry.get_variable_or_parameter(name) is None
    p = inf.Parameter(0, name=name)

    # the parameter exists in the context
    # randvar_registry.register_variable(x) has been automatically called
    assert len(randvar_registry.get_var_parameters()) == 1
    assert randvar_registry.get_var_parameters()[name] == p
    assert randvar_registry.get_variable_or_parameter(name) == p

    # if create a new parameter with the same name, it fails just if is_default is False
    if is_default:
        p = inf.Parameter(0, name=name)
        assert len(randvar_registry.get_var_parameters()) == 1
        assert randvar_registry.get_var_parameters()[name] == p
        assert randvar_registry.get_variable_or_parameter(name) == p
    else:
        with pytest.raises(ValueError):
            inf.Parameter(0, name=name)


def test_graph(init_context):
    # this test check if calling to update_graph update the graph or not depending on is_building
    is_building = init_context['is_building']
    elems = []  # we use this variable to append the expected elements in the graph

    # the variable does not exist in the context
    assert len(randvar_registry.get_graph()) == len(elems)
    inf.Normal(0, 1, name='x')
    randvar_registry.update_graph()
    if is_building:
        elems.append('x')

    # the variable exists in the context
    # randvar_registry.register_variable(x) has been automatically called
    assert len(randvar_registry.get_graph()) == len(elems)
    for elem in elems:
        assert elem in randvar_registry.get_graph()

    inf.Normal(0, 1, name='y')
    randvar_registry.update_graph()
    if is_building:
        elems.append('y')

    # the variable exists in the context
    # randvar_registry.register_variable(x) has been automatically called
    assert len(randvar_registry.get_graph()) == len(elems)
    for elem in elems:
        assert elem in randvar_registry.get_graph()



import pytest

import inferpy as inf
from inferpy.contextmanager import data_model


@pytest.mark.parametrize("size", [
    1,
    10,
    1000
])
def test_datamodel(size):
    with data_model.datamodel(size):
        x = inf.Normal(0, 1, name='x')
    assert x.sample_shape == size


@pytest.mark.parametrize("size", [
    1,
    10,
    1000
])
def test_fit(size):
    with data_model.fit(size):
        with data_model.datamodel():
            x = inf.Normal(0, 1, name='x')
    assert x.sample_shape == size


def test_is_active():
    assert not data_model.is_active()
    with data_model.datamodel():
        assert data_model.is_active()
    assert not data_model.is_active()


def test_has_datamodel_var_parameter():
    x = inf.Normal(0, 1)
    with data_model.datamodel(size=10):
        y = inf.Normal(x, 1)
        z = inf.Normal(y, 1)

    # uses the default random variable registry
    assert not data_model._has_datamodel_var_parameters(x.name)  # outside datamodel
    assert not data_model._has_datamodel_var_parameters(y.name)  # first level in datamodel (not var param)
    assert data_model._has_datamodel_var_parameters(z.name)  # has the y rand var as parameter




import tensorflow as tf
import warnings


"""
Module to manage global sessions and graphs executed in sessions
"""

__session = None


def new_session(gpu_memory_fraction=0.0):
    # Create a new session. By default do not use GPU. Use gpu_memory_fraction > 0 (and <= 1) to use GPU.
    if gpu_memory_fraction <= 0.0:
        set_session(tf.Session())
    else:
        config = tf.ConfigProto(log_device_placement=True)
        config.gpu_options.per_process_gpu_memory_fraction = gpu_memory_fraction  # memory_fraction [0-1]
        set_session(tf.Session(config=config))


def get_session():
    global __session
    if not __session:
        __session = tf.Session()
    return __session


def set_session(session):
    global __session
    if __session:
        warnings.warn("Running session closed to use the provided session instead")
        __session.close()
    __session = session


def swap_session(new_session):
    global __session
    old_session = __session
    __session = new_session
    return old_session


def clear_session():
    global __session
    if __session:
        __session.close()
    __session = tf.Session()



import tensorflow as tf
from tensorflow_probability import edward2 as ed
import numpy as np
import pytest

import inferpy as inf
from inferpy.models import random_variable


# TODO: boolean operators and iter are not tested
@pytest.mark.parametrize("tensor, expected", [
    # sum
    ("inf.Normal([0., 1.], 0) + inf.Normal([1., 2.], 0)", [1., 3.]),
    ("[0, 1] + inf.Normal([1., 2.], 0)", [1., 3.]),
    ("inf.Normal([0., 1.], 0) + [1, 2]", [1., 3.]),
    # sub
    ("inf.Normal([0., 1.], 0) - inf.Normal([1., 2.], 0)", [-1., -1.]),
    ("[0, 1] - inf.Normal([1., 2.], 0)", [-1., -1.]),
    ("inf.Normal([0., 1.], 0) - [1, 2]", [-1., -1.]),
    # mult
    ("inf.Normal([0., 1.], 0) * inf.Normal([1., 2.], 0)", [0., 2.]),
    ("[0, 1] * inf.Normal([1., 2.], 0)", [0., 2.]),
    ("inf.Normal([0., 1.], 0) * [1, 2]", [0., 2.]),
    # div
    ("inf.Normal([0., 1.], 0) / inf.Normal([1., 2.], 0)", [0., 0.5]),
    ("[0, 1] / inf.Normal([1., 2.], 0)", [0., 0.5]),
    ("inf.Normal([0., 1.], 0) / [1, 2]", [0., 0.5]),
    # mod
    ("inf.Normal([0., 1.], 0) % inf.Normal([2., 2.], 0)", [0., 1.]),
    ("[0, 1] % inf.Normal([2., 2.], 0)", [0., 1.]),
    ("inf.Normal([0., 1.], 0) % [2, 2]", [0., 1.]),
    # lt
    ("inf.Normal(0., 0) < inf.Normal(2., 0)", True),
    ("1 < inf.Normal(0., 0)", False),
    ("inf.Normal(1., 0) < 2", True),
    # le
    ("inf.Normal(2., 0) <= inf.Normal(2., 0)", True),
    ("1 <= inf.Normal(0., 0)", False),
    ("inf.Normal(1., 0) <= 1", True),
    # gt
    ("inf.Normal(5., 0) > inf.Normal(2., 0)", True),
    ("1 > inf.Normal(3., 0)", False),
    ("inf.Normal(3., 0) > 1", True),
    # getitem
    ("inf.Normal([0., 1., 2., 3.], 0)[0]", 0),
    ("inf.Normal([0., 1., 2., 3.], 0)[2]", 2),
    ("inf.Normal(tf.ones((3, 2)), 0)[2][0]", 1),
    # pow
    ("inf.Normal(2., 0) ** inf.Normal(3., 0)", 8.),
    ("3 ** inf.Normal(2., 0)", 9.),
    ("inf.Normal(3., 0) ** 3", 27.),
    # neg
    ("-inf.Normal(5., 0)", -5),
    ("-inf.Normal(-5., 0)", 5),
    # abs
    ("abs(inf.Normal(5., 0))", 5),
    ("abs(inf.Normal(-5., 0))", 5),
    # matmul
    ("tf.matmul(inf.Normal(tf.ones((2, 3), dtype=np.float32), 0), inf.Normal(tf.eye(3, dtype=np.float32), 0))",
     np.ones((2, 3))),
    ("tf.matmul(np.ones((2, 3), dtype=np.float32), inf.Normal(tf.eye(3, dtype=np.float32), 0))",
     np.ones((2, 3))),
    ("tf.matmul(inf.Normal(np.ones((2, 3), dtype=np.float32), 0), np.eye(3, dtype=np.float32))",
     np.ones((2, 3))),
])
def test_operations(tensor, expected):
    result = inf.get_session().run(eval(tensor))
    # assert that it is equal to expected
    assert np.array_equal(result, expected)


@pytest.mark.parametrize("model_object", [
    # Simple random variable using scalars as parameters
    (inf.Normal(0, 1)),
    # Simple random variable using a list as parameter
    (inf.Normal([0., 0., 0., 0.], 1)),
    # Simple random variable using a numpy array as parameter
    (inf.Normal(np.zeros(5), 1)),
    # Simple random variable using a tensor as parameter
    (inf.Normal(0, tf.ones(5))),
    # Simple random variable using another random variable as parameter
    (inf.Normal(inf.Normal(0, 1), 1)),
    # Simple random variable using a combination of the previously tested options as parameter
    (inf.Normal([inf.Normal(0, 1), 1., tf.constant(1.)], 1.)),
    # Random variable operation used to define a Random Variable
    (inf.Normal(inf.Normal(0, 1) + inf.Normal(0, 1), 1)),
])
def test_edward_type(model_object):
    assert isinstance(model_object.var, ed.RandomVariable)


def test_name():
    x = inf.Normal(0, 1, name='foo')
    assert x.name == 'foo'

    # using the name, not the tensor name
    x = inf.Normal(0, 1, name='foo')
    assert x.name == 'foo'

    # Automatic name generation. It starts with 'randvar_X', where initially X is 0
    x = inf.Normal(0, 1)
    assert isinstance(x.name, str)
    assert x.name == 'randvar_0'


def test_tensor_register():
    # This allows to run a inferpy.inf.RandomVariable directly in a tf session.

    x = inf.Normal(5., 0., name='foo')

    assert inf.get_session().run(x) == 5.
    assert isinstance(tf.convert_to_tensor(x), tf.Tensor)
    assert inf.get_session().run(tf.convert_to_tensor(x)) == 5.
    assert inf.get_session().run(tf.constant(5.) + x) == 10.
    assert inf.get_session().run(x + tf.constant(5.)) == 10.


def test_convert_random_variables_to_tensors():
    # element without RVs
    element = 1
    element = random_variable._convert_random_variables_to_tensors(element)
    assert element == element

    # list of elements different from RVs
    element = [1, 1]
    element = random_variable._convert_random_variables_to_tensors(element)
    assert element == element

    # numpy array of elements different from RVs
    element = np.ones((3, 2))
    element = random_variable._convert_random_variables_to_tensors(element)
    assert (element == element).all()

    # A single Random Variable
    element = inf.Normal(0, 1)
    result = random_variable._convert_random_variables_to_tensors(element)
    assert isinstance(element, random_variable.RandomVariable) and not isinstance(result, random_variable.RandomVariable)

    # A list with some Random Variables
    element = [inf.Normal(0, 1), 1, inf.Normal(0, 1), 2]
    result = random_variable._convert_random_variables_to_tensors(element)
    assert all([
        isinstance(element[i], random_variable.RandomVariable) and
        not isinstance(result[i], random_variable.RandomVariable)
        for i in [0, 2]])

    # A list with some nested Random Variables
    element = [[inf.Normal(0, 1), 1, inf.Normal(0, 1), 2]]
    result = random_variable._convert_random_variables_to_tensors(element)
    assert all([
        isinstance(element[0][i], random_variable.RandomVariable) and
        not isinstance(result[0][i], random_variable.RandomVariable)
        for i in [0, 2]])


def test_random_variable_in_pmodel():
    # test that random variables in pmodel works even if no name has been provided
    @inf.probmodel
    def model():
        inf.Normal(0, 1)

    v = list(model().vars.values())[0]
    assert v.name.startswith('randvar')
    # assert also is_datamodel is false
    assert not v.is_datamodel


def test_random_variable_in_datamodel():
    # test that random variables in datamodel which uses sample_shape warns about that, and uses datamodel size
    with pytest.warns(UserWarning):
        with inf.datamodel(10):
            x = inf.Normal(0, 1, sample_shape=(2,))

        assert x.sample_shape == 10
    # assert also that is_datamodel is true
    assert x.is_datamodel


def test_run_in_session():
    inf.get_session()
    x = inf.Normal(1, 0)
    assert inf.get_session().run(x) == 1



from contextlib import contextmanager
from inferpy.util import tf_graph
import warnings


# This dict store the context parameters of random variable and parameters registry; if the graph is being built or not;
# the graph of dependencies (can be built here or provided through the init function), and the builder vars and params
# that are being built now.

# Finally, there is a special registry which is the default (marked with `is_default`). The default registry is used to
# declare variables and parameters just to play, and therefore it is allowed to declare these elements using already
# declared names (overwriting theme). If it is the case, a warning message is raised.

def restart_default():
    global _default_properties
    _default_properties = dict(
        build_graph=True,
        graph=tf_graph.get_empty_graph(),
        builder_vars=dict(),
        builder_params=dict(),
        is_default=True
    )


# call to the function to restart the default context
restart_default()


# initially, _properties are the default properties
_properties = _default_properties


def is_building_graph():
    return _properties['build_graph']


def is_default():
    return _properties['is_default']


def register_variable(rv):
    # rv is a Random Variable from inferpy
    # the same name cannot be used in builder_vars and builder_params because names are used directly as graph nodes
    if rv.name in _properties['builder_vars'] or rv.name in _properties['builder_params']:
        if is_default() and rv.name not in _properties['builder_params']:
            # in default context; delete variable from builder_vars and graph to add the new one after removal
            del _properties['builder_vars'][rv.name]
            # if update_graph was called and rv name was included in graph, remove it too
            if rv.name in _properties['graph']:
                _properties['graph'].remove_node(rv.name)
            warnings.warn("The variable {} was already defined in the default random variable registry, \
                and is going to be removed. ".format(rv.name))
        else:
            raise ValueError('Random Variable names must be unique among Random Variables and Parameters. \
                             Detected twice: {}'.format(rv.name))
    _properties['builder_vars'][rv.name] = rv


def register_parameter(p):
    # p is a Parameter from inferpy
    # the same name cannot be used in builder_vars and builder_params because names are used directly as graph nodes
    if p.name in _properties['builder_params'] or p.name in _properties['builder_vars']:
        if is_default() and p.name not in _properties['builder_vars']:
            # in default context; delete parameter from builder_params and graph to add the new one after removal
            del _properties['builder_params'][p.name]
            # if update_graph was called and parameter name was included in graph, remove it too
            if p.name in _properties['graph']:
                _properties['graph'].remove_node(p.name)
            warnings.warn("The parameter {} was already defined in the default random parameter registry, \
                and is going to be removed. ".format(p.name))
        else:
            raise ValueError('Parameter names must be unique among Parameters and Random Variables. \
                             Detected twice: {}'.format(p.name))
    _properties['builder_params'][p.name] = p


def get_variable(name):
    return _properties['builder_vars'].get(name, None)


def get_variable_or_parameter(name):
    # return the variable or parameter if exists. Otherwise, return None
    return _properties['builder_vars'].get(
        name,
        _properties['builder_params'].get(name, None)
        )


def get_var_parameters():
    # return a copy of the internal dict properties field 'builder_params', just to
    # avoid the modification of the _properties dict from outside
    return {k: p for k, p in _properties['builder_params'].items()}


def get_graph():
    # return the graph of dependencies of the prob model that is being built
    return _properties['graph']


def update_graph(rv_name=None):
    # update the graph by creating a new one using the actual tf computational graph
    # it uses the actual random variables and parameters, and the rv_name if exists
    # only updates the model if the property build_graph is True
    if _properties['build_graph']:
        # compute all the desired names in the graph (only inferpy RandomVariable and Parameters, and ed2.RandomVariable)
        elements_set = set(_properties['builder_vars']).union(
                           set(_properties['builder_params'])
                          )
        # if rv_name, use this name too
        if rv_name:
            elements_set.add(rv_name)
        # now create the dependencies graph
        _properties['graph'] = tf_graph.get_graph(elements_set)


@contextmanager
def init(graph=None):
    global _properties
    # random variable and parameters registry context. Allows to get access to RVs and parameters as they are built
    # (at the same time ed.tape registers vars)
    # We only allow to use one context level, checked by is_default field (if false, init was called and not exit before)
    assert _properties['is_default']

    # create a new dict, so the default dict is not modified
    _properties = dict()
    _properties['is_default'] = False

    # if graph is not None, use this element as the graph, and do not update it in this context
    _properties['build_graph'] = graph is None
    # if graph is none, start from an empty graph
    if _properties['build_graph']:
        _properties['graph'] = tf_graph.get_empty_graph()
    else:
        _properties['graph'] = graph
    # random variables and parameter dict registry are initially empty
    _properties['builder_vars'] = dict()
    _properties['builder_params'] = dict()
    try:
        yield
    finally:
        # reasign the default registry
        _properties = _default_properties



import importlib
import pytest
import numpy as np
import random
import tensorflow as tf

import inferpy as inf
from inferpy.contextmanager import randvar_registry
from inferpy.util import name


SEED = 607742924


@pytest.fixture(autouse=True)
def tf_new_default_graph():
    """
    Reset the tf default graph at teardown, so parametrize can declare tensors
    """
    g = tf.Graph()
    with g.as_default():
        with tf.Session() as sess:
            inf.set_session(sess)
            yield g


@pytest.fixture(autouse=True)
def reproducible():
    """
    Set tf, numpy and random seed to SEED
    """
    tf.random.set_random_seed(SEED)
    np.random.seed(SEED)
    random.seed(SEED)
    yield


@pytest.fixture(autouse=True)
def restart_default_randvar_registry():
    randvar_registry.restart_default()
    yield


@pytest.fixture(autouse=True)
def restart_random_names_counter():
    importlib.reload(name)
    yield



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import functools
from enum import IntEnum
import tensorflow as tf
from tensorflow_probability import edward2 as ed
from tensorflow_probability.python.edward2.interceptor import interceptable

import tensorflow_probability as tfp
from tensorflow_probability.python.edward2 import generated_random_variables
from tensorflow.python.client import session as tf_session
import warnings

from inferpy import contextmanager
from inferpy import util
import inferpy.util.session


# the list of available RandomVariables in edward2. Matches with the available distributions in tensorflow_probability
distributions_all = [rv for rv in generated_random_variables.__all__ if rv != 'as_random_variable']


def _make_edward_random_variable(distribution_obj):
    """Factory function to make random variable given distribution class."""

    @interceptable
    @functools.wraps(distribution_obj, assigned=('__module__', '__name__'))
    def func(*args, **kwargs):
        # pylint: disable=g-doc-args
        """Create a random variable for ${cls}.
        See ${cls} for more details.
        Returns:
          RandomVariable.
        #### Original Docstring for Distribution
        ${doc}
        """
        # pylint: enable=g-doc-args
        sample_shape = kwargs.pop('sample_shape', ())
        value = kwargs.pop('value', None)
        return ed.RandomVariable(distribution=distribution_obj,
                                 sample_shape=sample_shape,
                                 value=value)
    return func


def _try_sess_run(p, sess):
    try:
        return sess.run(p)
    except (RuntimeError, TypeError, ValueError):
        return p


class Kind(IntEnum):
    GLOBAL_HIDDEN = 0
    GLOBAL_OBSERVED = 1
    LOCAL_HIDDEN = 2
    LOCAL_OBSERVED = 3


class RandomVariable:
    """
    Class for random variables. It encapsulates the Random Variable from edward2, and additional properties.

    - It creates a variable generator. It must be a function without parameters, that creates a
      new Random Variable from edward2. It is used to define edward2 models as functions.
      Also, it is useful to define models using the intercept function.

    - The first time the var property is used, it creates a var using the variable generator.
    """

    def __init__(self, var, name, is_datamodel, ed_cls, var_args, var_kwargs, sample_shape,
                 is_observed, observed_value):
        self.var = var
        self.is_datamodel = is_datamodel
        # These parameters are used to allow the re-creation of the random var by build_in_session function
        self._ed_cls = ed_cls
        self._var_args = var_args
        self._var_kwargs = var_kwargs
        self._sample_shape = sample_shape
        self.is_observed = is_observed
        self.observed_value = observed_value

        # if name is provided, use it. Otherwise, use it from var or var.distribution
        if name is None:
            self.name = self.__getattr__('name')
        else:
            self.name = name

    @property
    def type(self):
        first_part = 'LOCAL' if self.is_datamodel else 'GLOBAL'
        second_part = 'OBSERVED' if util.get_session().run(self.is_observed) else 'HIDDEN'

        return getattr(Kind, "{}_{}".format(first_part, second_part))

    def build_in_session(self, sess):
        """
        Allow to build a copy of the random variable but running previously each parameter in the tf session.
        This way, it uses the value of each tf variable or placeholder as a tensor, not as a tf variable or placeholder.
        If this random variable is a ed random variable directly assigned to .var, we cannot re-create it. In this
        case, return self.
        :param sess: tf session used to run each parameter used to build this random variable.
        :returns: the random variable object
        """
        # Cannot re-create the random variable. Return this var itself
        if self._ed_cls is None:
            return self

        # create the ed random variable evaluating each parameter in a tf session
        ed_random_var = self._ed_cls(*[_try_sess_run(a, sess) for a in self._var_args],
                                     **{k: _try_sess_run(v, sess) for k, v in self._var_kwargs.items()},
                                     sample_shape=self._sample_shape)

        initial_value = util.get_session().run(self.observed_value_var)
        is_observed, observed_value = _make_predictable_variables(initial_value, self.name)
        # build the random variable by using the ed random var
        rv = RandomVariable(
            var=ed_random_var,
            name=self.name,
            is_datamodel=self.is_datamodel,
            ed_cls=self._ed_cls,
            var_args=self._var_args,
            var_kwargs=self._var_kwargs,
            sample_shape=self._sample_shape,
            is_observed=self.is_observed,
            observed_value=self.observed_value
        )

        # put the docstring and the name as well as in _make_random_variable function
        docs = RandomVariable.__doc__ + '\n Random Variable information:\n' + \
            ('-' * 30) + '\n' + self._ed_cls.__doc__
        name = self._ed_cls.__name__

        rv.__doc__ += docs
        rv.__name__ = name

        return rv

    def copy(self):
        """
        Makes a of the current random variable where the distribution parameters are fixed.
        :return: new object of class RandomVariable
        """
        return self.build_in_session(inferpy.get_session())

    def __repr__(self):
        # Custom representation of the random variable
        string = "inf.RandomVariable ({} distribution) named {}, shape={}, dtype={}".format(
            self.__name__, self.distribution.name, self.shape, self.dtype.name)

        return "<%s>" % string

    def __getattr__(self, name):
        """
        If try to use attributes or functions not defined for RandomVariables, this function is executed.
        First try to return the same attribute of function from the edward2 RandomVariable.
        Secondly try to return the same element from the distribution object inside the edward2 RandomVariable.
        Otherwise, raise an Exception.
        """
        if hasattr(self.var, name):
            obj = getattr(self.var, name)
            if hasattr(obj, '__call__'):
                return util.tf_run_allowed(obj)
            else:
                return obj
        elif hasattr(self.var.distribution, name):
            obj = getattr(self.var.distribution, name)
            if hasattr(obj, '__call__'):
                return util.tf_run_allowed(obj)
            else:
                return obj
        else:
            raise AttributeError('Property or method "{name}" not implemented in "{classname}"'.format(
                name=name,
                classname=type(self.var).__name__
            ))

    # In the following, define the operations by using the edward2 operator definition

    def __add__(self, other):
        return _operator(self.var, other, '__add__')

    def __radd__(self, other):
        return _operator(self.var, other, '__radd__')

    def __sub__(self, other):
        return _operator(self.var, other, '__sub__')

    def __rsub__(self, other):
        return _operator(self.var, other, '__rsub__')

    def __mul__(self, other):
        return _operator(self.var, other, '__mul__')

    def __rmul__(self, other):
        return _operator(self.var, other, '__rmul__')

    def __div__(self, other):
        return _operator(self.var, other, '__div__')

    def __rdiv__(self, other):
        return _operator(self.var, other, '__rdiv__')

    def __truediv__(self, other):
        return _operator(self.var, other, '__truediv__')

    def __rtruediv__(self, other):
        return _operator(self.var, other, '__rtruediv__')

    def __floordiv__(self, other):
        return _operator(self.var, other, '__floordiv__')

    def __rfloordiv__(self, other):
        return _operator(self.var, other, '__rfloordiv__')

    def __mod__(self, other):
        return _operator(self.var, other, '__mod__')

    def __rmod__(self, other):
        return _operator(self.var, other, '__rmod__')

    def __lt__(self, other):
        return _operator(self.var, other, '__lt__')

    def __le__(self, other):
        return _operator(self.var, other, '__le__')

    def __gt__(self, other):
        return _operator(self.var, other, '__gt__')

    def __ge__(self, other):
        return _operator(self.var, other, '__ge__')

    def __and__(self, other):
        return _operator(self.var, other, '__and__')

    def __rand__(self, other):
        return _operator(self.var, other, '__rand__')

    def __or__(self, other):
        return _operator(self.var, other, '__or__')

    def __ror__(self, other):
        return _operator(self.var, other, '__ror__')

    def __xor__(self, other):
        return _operator(self.var, other, '__xor__')

    def __rxor__(self, other):
        return _operator(self.var, other, '__rxor__')

    def __getitem__(self, *args):
        return self.var.__getitem__(*args)

    def __pow__(self, other):
        return _operator(self.var, other, '__pow__')

    def __rpow__(self, other):
        return _operator(self.var, other, '__rpow__')

    def __invert__(self):
        return self.var.__invert__()

    def __neg__(self):
        return self.var.__neg__()

    def __abs__(self):
        return self.var.__abs__()

    def __matmul__(self, other):
        return _operator(self.var, other, '__matmul__')

    def __rmatmul__(self, other):
        return _operator(self.var, other, '__rmatmul__')

    def __iter__(self):
        return self.var.__iter__()

    def __bool__(self):
        return self.var.__bool__()

    def __nonzero__(self):
        return self.var.__nonzero__()


def _convert_random_variables_to_tensors(arg):
    # function used to convert Random Variables (from inferpy or edward2) to tensors, even if there
    # are in a list or nested list, in order to allow to use them as arguments for other Random Variables
    if isinstance(arg, (ed.RandomVariable, RandomVariable)):
        return tf.convert_to_tensor(arg)
    elif isinstance(arg, list):
        return [_convert_random_variables_to_tensors(nested_arg) for nested_arg in arg]
    else:
        return arg


def _make_random_variable(distribution_name):
    """Factory function to make random variable given distribution name."""

    distribution_cls = getattr(tfp.distributions, distribution_name)
    ed_random_variable_cls = getattr(ed, distribution_name)

    docs = RandomVariable.__doc__ + '\n Random Variable information:\n' + \
        ('-' * 30) + '\n' + ed_random_variable_cls.__doc__
    name = ed_random_variable_cls.__name__

    def func(*args, **kwargs):
        # The name used to identify the random variable by string
        if 'name' not in kwargs:
            kwargs['name'] = util.name.generate('randvar')
        rv_name = kwargs.get('name')

        if contextmanager.data_model.is_active():
            if 'sample_shape' in kwargs:
                # warn that sampe_shape will be ignored
                warnings.warn('Random Variables defined inside a probabilistic model ignore the sample_shape argument.')
                kwargs.pop('sample_shape', None)
            sample_shape = ()  # used in case that RV is in probmodel, but not in a datamodel
        else:
            # only used if prob model is active
            sample_shape = kwargs.pop('sample_shape', ())

        # convert any Random Variable in a list or nested list in arguments to tensors, allowing to use RV's as arguments
        sanitized_args = [_convert_random_variables_to_tensors(arg) for arg in args]
        sanitized_kwargs = {k: _convert_random_variables_to_tensors(v) for k, v in kwargs.items()}

        # If it is inside a data model, ommit the sample_shape in kwargs if exist and use size from data_model
        # NOTE: Needed here because we need to know the shape of the distribution, as well as its dtype
        # Not using sample shape yet. Used just to create the tensors, and compute the dependencies using the tf graph
        tfp_dist = distribution_cls(*sanitized_args, **sanitized_kwargs)
        if contextmanager.data_model.is_active():
            # create graph once tensors are registered in graph
            contextmanager.randvar_registry.update_graph(rv_name)

            # compute sample_shape now that we have computed the dependencies
            sample_shape = contextmanager.data_model.get_sample_shape(rv_name)

            # create tf.Variable's to allow to observe the Random Variable
            shape = ([sample_shape] if sample_shape else []) + \
                tfp_dist.batch_shape.as_list() + \
                tfp_dist.event_shape.as_list()

            # take into account the dtype of tfp_dist in order to create the initial value correctly
            initial_value = tf.zeros(shape, dtype=tfp_dist.dtype) if shape else tf.constant(0,  dtype=tfp_dist.dtype)

            # build the respective boolean and tf.Variables
            is_observed, observed_value = _make_predictable_variables(initial_value, rv_name)

            # use this context to intercept the value using the tf.cond dependent on the previous tf.Variables
            with ed.interception(util.interceptor.set_values_condition(is_observed, observed_value)):
                ed_random_var = _make_edward_random_variable(tfp_dist)(sample_shape=sample_shape, name=rv_name)

            # if the random variable has been intercepted with other random variable, which also has the tf.Variables,
            # then use their tf.Variables instead of the previously created tf.Variables
            if hasattr(ed_random_var, 'is_observed') and hasattr(ed_random_var, 'observed_value'):
                is_observed = ed_random_var.is_observed
                observed_value = ed_random_var.observed_value

            is_datamodel = True
        else:
            # create tf.Variable's to allow to observe the Random Variable
            shape = tfp_dist.batch_shape.as_list() + tfp_dist.event_shape.as_list()

            # take into account the dtype of tfp_dist in order to create the initial value correctly
            initial_value = tf.zeros(shape, dtype=tfp_dist.dtype) if shape else tf.constant(0,  dtype=tfp_dist.dtype)

            # build the respective boolean and tf.Variables
            is_observed, observed_value = _make_predictable_variables(initial_value, rv_name)

            # use this context to intercept the value using the tf.cond dependent on the previous tf.Variables
            with ed.interception(util.interceptor.set_values_condition(is_observed, observed_value)):
                # sample_shape is sample_shape in kwargs or ()
                ed_random_var = ed_random_variable_cls(*sanitized_args, **sanitized_kwargs, sample_shape=sample_shape)

            # if the random variable has been intercepted with other random variable, which also has the tf.Variables,
            # then use their tf.Variables instead of the previously created tf.Variables
            if hasattr(ed_random_var, 'is_observed') and hasattr(ed_random_var, 'observed_value'):
                is_observed = ed_random_var.is_observed
                observed_value = ed_random_var.observed_value

            is_datamodel = False

        rv = RandomVariable(
            var=ed_random_var,
            name=rv_name,
            is_datamodel=is_datamodel,
            ed_cls=ed_random_variable_cls,
            var_args=sanitized_args,
            var_kwargs=sanitized_kwargs,
            sample_shape=sample_shape,
            is_observed=is_observed,
            observed_value=observed_value,
        )

        # register the variable as it is created. Used to detect dependencies
        contextmanager.randvar_registry.register_variable(rv)
        contextmanager.randvar_registry.update_graph()

        # Doc for help menu
        rv.__doc__ += docs
        rv.__name__ = name

        return rv

    # Doc for help menu
    func.__doc__ = docs
    func.__name__ = name
    return func


def _make_predictable_variables(initial_value, rv_name):
    is_observed = tf.Variable(False, trainable=False,
                              name="inferpy-predict-enabled-{name}".format(name=rv_name or "default"))

    observed_value = tf.Variable(initial_value, trainable=False,
                                 name="inferpy-predict-{name}".format(name=rv_name or "default"))

    util.session.get_session().run(tf.variables_initializer([is_observed, observed_value]))

    return is_observed, observed_value


def _operator(var, other, operator_name):
    """
    Function to apply the operation called `operator_name` by using a RandomVariable
    object and other object (which can be a RandomVariable or not).
    """
    # get the operator_name function from the edward2 variable `var`
    # and call that function using the edward2 object if other is a RandomVariable,
    # otherwise just use `other` as argument
    return getattr(var, operator_name)(other.var if isinstance(other, RandomVariable) else other)


# Define all Random Variables existing in edward2
for d in distributions_all:
    globals()[d] = _make_random_variable(d)


def _tensor_conversion_function(rv, dtype=None, name=None, as_ref=False):
    """
        Function that converts the inferpy variable into a Tensor.
        This will enable the use of enable tf.convert_to_tensor(rv)

        If the variable needs to be broadcast_to, do it right now
    """
    # return the tf.Variable last snapshot if it is observed, and the ed2 evaluation (ed2.value) otherwise
    return tf.convert_to_tensor(rv.observed_value.value() if util.get_session().run(rv.is_observed) else rv.var)


# register the conversion function into a tensor
tf.register_tensor_conversion_function(  # enable tf.convert_to_tensor
    RandomVariable, _tensor_conversion_function)


def _session_run_conversion_fetch_function(rv):
    """
        This will enable run and operations with other tensors
    """
    return ([tf.convert_to_tensor(rv)], lambda val: val[0])


tf_session.register_session_run_conversion_functions(  # enable sess.run, eval
    RandomVariable,
    _session_run_conversion_fetch_function)



import numpy as np
import inferpy as inf
N=1000
x_train = np.concatenate([inf.Normal([0.0,0.0], scale=1.).sample(int(N/2)), inf.Normal([10.0,10.0], scale=1.).sample(int(N/2))])
x_test = np.concatenate([inf.Normal([0.0,0.0], scale=1.).sample(int(N/2)), inf.Normal([10.0,10.0], scale=1.).sample(int(N/2))])

##########

import inferpy as inf
import tensorflow as tf

# definition of a generic model


# number of components
k = 1
# size of the hidden layer in the NN
d0 = 100
# dimensionality of the data
dx = 2
# number of observations (dataset size)
N = 1000


@inf.probmodel
def nlpca(k, d0, dx, decoder):

    with inf.datamodel():
        z = inf.Normal(tf.ones([k])*0.5, 1., name="z")    # shape = [N,k]
        output = decoder(z,d0,dx)
        x_loc = output[:,:dx]
        x_scale = tf.nn.softmax(output[:,dx:])
        x = inf.Normal(x_loc, x_scale, name="x")   # shape = [N,d]


def decoder(z,d0,dx):
    h0 = tf.layers.dense(z, d0, tf.nn.relu)
    return tf.layers.dense(h0, 2 * dx)


# Q-model  approximating P

@inf.probmodel
def qmodel(k):
    with inf.datamodel():
        qz_loc = inf.Parameter(tf.ones([k])*0.5, name="qz_loc")
        qz_scale = tf.math.softplus(inf.Parameter(tf.ones([k]),name="qz_scale"))

        qz = inf.Normal(qz_loc, qz_scale, name="z")


# create an instance of the model
m = nlpca(k,d0,dx, decoder)

# set the inference algorithm
VI = inf.inference.VI(qmodel(k), epochs=5000)

# learn the parameters
m.fit({"x": x_train}, VI)

# extract the hidden encoding
hidden_encoding = m.posterior("z").parameters()["loc"]

# project x_test into the reduced space (encode)
m.posterior("z", data={"x": x_test}).sample(5)

# sample from the posterior predictive (i.e., simulate values for x given the learnt hidden)
m.posterior_predictive("x").sample(5)

# decode values from the hidden representation
m.posterior_predictive("x", data={"z": [2]}).sample(5)




from contextlib import contextmanager, ExitStack
from . import randvar_registry


# This dict store the active (if active is True) context and the size of the dtamodel plate in the current context
# The datamodel means that the variables inside are expanded. At least, the size is equals 1.
_active_datamodel = dict(
    size=1,
    active=False
)


def is_active():
    # is the data model context active?
    return _active_datamodel['active']


def _has_datamodel_var_parameters(name):
    graph = randvar_registry.get_graph()
    # is this a Random Variable with any parent expanded? If any, return True (will be expanded by parent size)
    # To do that we check the is_datamodel from the parents (predecessors), which can be variables or parameters
    return any(randvar_registry.get_variable_or_parameter(pname).is_datamodel for pname in graph.predecessors(name))


def get_sample_shape(name):
    """
    This function must be used inside a datamodel context (it is not checked here)
    If var parameters are not expanded, then expand.
        :name (str): The name of the variable to get its sample shape
        :returns: a the sample_shape (number of samples of the datamodel). It is an integer, or ().
    """

    # Parameters already expanded? (remember that in probmodel definitions, each RandomVariable must have a name)
    if _has_datamodel_var_parameters(name):
        # yes, do not need to expand this var (it will be expanded by parents)
        size = ()
    else:
        # no, we need to expand this variable
        size = _active_datamodel['size']

    return size


@contextmanager
def fit(size):
    # size must be an integer
    if not isinstance(size, int):
        raise TypeError('The size of the data model must be an integer, not : {}'.format(type(size)))
    # Fit the datamodel parameters
    _active_datamodel['size'] = size

    try:
        yield
    finally:
        _active_datamodel['size'] = 1


@contextmanager
def datamodel(size=None):
    """
    This context is used to declare a plateau model. Random Variables and Parameters will use a sample_shape
    defined by the argument `size`, or by the `data_model.fit`. If `size` is not specify, the default size 1,
    or the size specified by `fit` will be used.
    """

    # NOTE: We only allow to use one context level, assert that it is not active now
    assert not _active_datamodel['active']
    _active_datamodel['active'] = True

    # to simplify the code, avoiding if-else blocks, we declare a list of contexts (empty or with one fit if `size`)
    # and use the ExitStack() to enter all of them at the same time (if empty, it does nothing)
    contexts = []
    # if size is provided, use the fit context
    if size:
        contexts.append(fit(size))

    # use the ExitStack to enter the context, or do nothing if contexts is empty
    try:
        with ExitStack() as stack:
            for c in contexts:
                stack.enter_context(c)
            yield
    finally:
        _active_datamodel['active'] = False



from collections import defaultdict

# for each used prefix, return a different counter starting from 0
prefixes_count = defaultdict(int)


def generate(prefix):
    """This function is used to generate names based on an incremental counter (global variable in this module)
        dependent on the prefix (staring from 0 index)

        :prefix (`str`): The begining of the random generated name

        :returns: The generated random name
    """
    name = "{}_{}".format(prefix, prefixes_count[prefix])
    prefixes_count[prefix] = prefixes_count[prefix] + 1
    return name



import inferpy as inf
import tensorflow as tf


@inf.probmodel
def linear_reg(d):
    w0 = inf.Normal(0, 1, name="w0")
    w = inf.Normal(tf.zeros([d,1]), 1, name="w")

    with inf.datamodel():
        x = inf.Normal(tf.ones([d]), 2, name="x")
        y = inf.Normal(w0 + x @ w, 1.0, name="y")


@inf.probmodel
def qmodel(d):
    qw0_loc = inf.Parameter(0., name="qw0_loc")
    qw0_scale = tf.math.softplus(inf.Parameter(1., name="qw0_scale"))
    qw0 = inf.Normal(qw0_loc, qw0_scale, name="w0")

    qw_loc = inf.Parameter(tf.zeros([d,1]), name="qw_loc")
    qw_scale = tf.math.softplus(inf.Parameter(tf.ones([d,1]), name="qw_scale"))
    qw = inf.Normal(qw_loc, qw_scale, name="w")


# create an instance of the model
m = linear_reg(d=2)

### create toy data
N = 1000
data = m.prior(["x", "y"], data={"w0":0, "w":[[2],[1]]}).sample(N)

x_train = data["x"]
y_train = data["y"]

# set and run the inference
VI = inf.inference.VI(qmodel(2), epochs=10000)
m.fit({"x": x_train, "y":y_train}, VI)


# extract the parameters of the posterior
m.posterior(["w", "w0"]).parameters()


import numpy as np
import inferpy as inf
N=1000
x_train = np.concatenate([inf.Normal([0.0,0.0], scale=1.).sample(int(N/2)), inf.Normal([10.0,10.0], scale=1.).sample(int(N/2))])
##########


# required pacakges
import inferpy as inf
import numpy as np
import tensorflow as tf

# definition of a generic model


# number of components
k = 1
# size of the hidden layer in the NN
d0 = 100
# dimensionality of the data
dx = 2
# number of observations (dataset size)
N = 1000


@inf.probmodel
def vae(k, d0, dx, decoder):

    with inf.datamodel():
        z = inf.Normal(tf.ones([k])*0.5, 1., name="z")    # shape = [N,k]
        output = decoder(z,d0,dx)
        x_loc = output[:,:dx]
        x_scale = tf.nn.softmax(output[:,dx:])
        x = inf.Normal(x_loc, x_scale, name="x")   # shape = [N,d]


def decoder(z,d0,dx):   # k -> d0 -> 2*dx
    h0 = tf.layers.dense(z, d0, tf.nn.relu)
    return tf.layers.dense(h0, 2 * dx)


# Q-model  approximating P
def encoder(x, d0, k): # dx -> d0 -> 2*k
    h0 = tf.layers.dense(tf.constant(x), d0, tf.nn.relu)
    return tf.layers.dense(h0, 2*k)

@inf.probmodel
def qmodel(k, d0, x, encoder):

    output = encoder(x, d0, k)
    qz_loc = output[:, :k]
    qz_scale = tf.nn.softmax(output[:, k:])

    qz = inf.Normal(qz_loc, qz_scale, name="z")


# create an instance of the model
m = vae(k,d0,dx, decoder)
q = qmodel(k,d0,x_train,encoder)

# set the inference algorithm
VI = inf.inference.VI(q, epochs=5000)

# learn the parameters
m.fit({"x": x_train}, VI)


#extract the hidden representation
hidden_encoding = m.posterior["z"]
sess = inf.get_session()
print(sess.run(hidden_encoding.loc))



import inferpy as inf


def test_parameter_in_pmodel():
    # test that random variables in pmodel works even if no name has been provided
    @inf.probmodel
    def model():
        inf.Parameter(0)

    v = list(model().params.values())[0]
    assert v.name.startswith('parameter')
    # assert also is_datamodel is false
    assert not v.is_datamodel


def test_parameter_in_datamodel():
    with inf.datamodel(10):
        x = inf.Parameter(0)

    # assert that is_datamodel is true
    assert x.is_datamodel


def test_run_in_session():
    x = inf.Parameter(0)
    assert inf.get_session().run(x) == 0



from inferpy.contextmanager import datamodel  # noqa: F401
from .random_variable import *  # noqa: F403
from .prob_model import probmodel  # noqa: F401
from .parameter import Parameter  # noqa: F401
from inferpy import inference  # noqa: F401


__all__ = [
            'datamodel',
            'inference',
            'Parameter',
            'probmodel',
        ] + distributions_all  # noqa: F405



import os
from os.path import isfile, join


# List of files that we do not want to test
BLACK_LIST = ["defmodels.py"]


def test_examples():
    pth = "examples_old/"
    inf_examples = [f for f in os.listdir(pth) if isfile(join(pth, f)) and
                    f.endswith(".py") and not f.endswith("__init__.py") and f not in BLACK_LIST]

    failed = []

    for f in inf_examples:
        filename = join(pth, f)
        print("testing " + filename)
        try:
            exec(compile(open(filename, "rb").read(), filename, 'exec'))
        except Exception as e:
            failed.append(f)
            print("ERROR:")
            print(e)

    if len(failed) > 0:
        print("failed:")
        print(failed)

    assert len(failed) == 0



import pytest
import tensorflow as tf

import inferpy as inf
from inferpy.util import runtime


@pytest.fixture(params=[True, False])
def is_tf_run_default(request):
    inf.set_tf_run(request.param)
    yield request.param


def test_tf_run_allowed(is_tf_run_default):
    # create a function which directly returns a tensor
    @runtime.tf_run_allowed
    def foo():
        return tf.ones(1)

    if is_tf_run_default:
        # in this case the tensor is run in a session
        assert foo() == [1.]
    else:
        # in this case the function returns a tensor
        assert isinstance(foo(), tf.Tensor)


def test_tf_run_allowed_nested(is_tf_run_default):
    # create a function which directly returns a tensor created by a second function
    # store the result of the nested call function bar
    res_bar = None

    @runtime.tf_run_allowed
    def foo():
        return bar()

    @runtime.tf_run_allowed
    def bar():
        nonlocal res_bar
        res_bar = tf.ones(1)
        return res_bar

    if is_tf_run_default:
        # in this case the tensor is run in a session
        assert foo() == [1.]
    else:
        # in this case the function returns a tensor
        assert isinstance(foo(), tf.Tensor)

    # independently of is_tf_run_default, res_bar is always a tensor
    assert isinstance(res_bar, tf.Tensor)


def test_tf_run_ignored(is_tf_run_default):
    # create a function which directly returns a tensor

    @runtime.tf_run_ignored
    def foo():
        return bar()

    @runtime.tf_run_allowed
    def bar():
        return tf.ones(1)

    # independently of is_tf_run_default, foo always returns a tensor
    assert isinstance(foo(), tf.Tensor)



import inferpy as inf
import numpy as np
import tensorflow as tf


@inf.probmodel
def pca(k,d):
    w = inf.Normal(loc=tf.zeros([k,d]), scale=1, name="w")      # shape = [k,d]
    with inf.datamodel():
        z = inf.Normal(tf.ones([k]),1, name="z")                # shape = [N,k]
        x = inf.Normal(z @ w , 1, name="x")                     # shape = [N,d]






#### data (do not show)
# generate training data
N = 1000
sess = tf.Session()
x_train = np.concatenate([inf.Normal([0.0,0.0], scale=1.).sample(int(N/2)), inf.Normal([10.0,10.0], scale=1.).sample(int(N/2))])



### encapsulated inference

#28


@inf.probmodel
def qmodel(k,d):
    qw_loc = inf.Parameter(tf.ones([k,d]), name="qw_loc")
    qw_scale = tf.math.softplus(inf.Parameter(tf.ones([k, d]), name="qw_scale"))
    qw = inf.Normal(qw_loc, qw_scale, name="w")

    with inf.datamodel():
        qz_loc = inf.Parameter(tf.ones([k]), name="qz_loc")
        qz_scale = tf.math.softplus(inf.Parameter(tf.ones([k]), name="qz_scale"))
        qz = inf.Normal(qz_loc, qz_scale, name="z")



# set the inference algorithm
VI = inf.inference.VI(qmodel(k=1,d=2), epochs=1000)
# create an instance of the model
m = pca(k=1,d=2)
# run the inference
m.fit({"x": x_train}, VI)

"""
 0 epochs	 44601.14453125....................
 200 epochs	 44196.98046875....................
 400 epochs	 50616.359375....................
 600 epochs	 41085.6484375....................
 800 epochs	 30349.79296875....................
 
"""

print(m.posterior("w").parameters())
"""
>>> m.posterior("w").parameters()
{'name': 'w',
 'allow_nan_stats': True,
 'validate_args': False,
 'scale': array([[0.9834974 , 0.99731755]], dtype=float32),
 'loc': array([[1.7543027, 1.7246702]], dtype=float32)}
"""

# 70


## ELBO definition


# define custom elbo function
def custom_elbo(pvars, qvars, **kwargs):

    # compute energy
    energy = tf.reduce_sum([tf.reduce_sum(p.log_prob(p.value)) for p in pvars.values()])

    # compute entropy
    entropy = - tf.reduce_sum([tf.reduce_sum(q.log_prob(q.value)) for q in qvars.values()])

    # compute ELBO
    ELBO = energy + entropy

    # This function will be minimized. Return minus ELBO
    return -ELBO



# set the inference algorithm
VI = inf.inference.VI(qmodel(k=1,d=2), loss=custom_elbo, epochs=1000)

# run the inference
m.fit({"x": x_train}, VI)




## 102


### optimization loop

# instance
q = qmodel(k=1,d=2)

# for not evaluating ELBO tensor
inf.util.runtime.set_tf_run(False)
# extract the computational graph of the ELBO
loss_tensor = inf.inference.loss_functions.elbo.ELBO(m,q, {"x": x_train})

# build an optimizer to minimize the ELBO
optimizer = tf.train.AdamOptimizer(learning_rate=0.1)
train = optimizer.minimize(loss_tensor)

# start a session
sess = tf.Session()
# intialize the TF variables
sess.run(tf.global_variables_initializer())



# optimization loop
t = []
for i in range(0,100):
    sess.run(train)
    t += [sess.run(loss_tensor)]
    print(t[-1])


# extract the posterior distributions
posterior_qvars = {name: qv.build_in_session(sess) for name, qv in q._last_expanded_vars.items()}





# 141


SVI = inf.inference.SVI(qmodel(k=1,d=2), epochs=1000, batch_size=200)




# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


"""
Module focused on evaluating tensors to makes the usage easier, forgetting about tensors and sessions
"""

from functools import wraps
from contextlib import contextmanager

from inferpy import util


# default value for tf_run in decorated tf_run_allowed functions
__tf_run_default = True

# configuration environment for runner_scopes. It counts the number of nested contexts
runner_context = dict(
    runner_recursive_depth=0
)


@contextmanager
def runner_scope():
    # Update the runner recursive depth, because decorated functions might call other decorated functions too.
    # This way, we can control that only first level decorated functions will be evaluated in a tf Session.
    runner_context['runner_recursive_depth'] += 1
    try:
        yield
    finally:
        runner_context['runner_recursive_depth'] -= 1


def tf_run_allowed(f):
    """
    A function might return a tensor or not. In order to decide if the result of this function needs to be evaluated
    in a tf session or not, use the tf_run extra parameter or the tf_run_default value. If True, and this function is
    in the first level of execution depth, use a tf Session to evaluate the tensor or other evaluable object (like dicts)
    """
    @wraps(f)
    def wrapper(*args, **kwargs):
        # first obtains the tf_run, a bool which tells if we need to eval the output in a session or not
        if "tf_run" in kwargs:
            tf_run = kwargs.pop("tf_run")
        else:
            tf_run = __tf_run_default

        # use this context to keep track of the decorated functions calls (recursive depth level)
        with runner_scope():
            # now execute the function
            obj = f(*args, **kwargs)
            if tf_run and runner_context['runner_recursive_depth'] == 1:
                # first recursive depth, and tf_run is True: we can eval the function
                return try_run(obj)
            else:
                # tf_run is False or we are in a deeper runner levels than 1 (do not eval the result yet)
                return obj
    return wrapper


def tf_run_ignored(f):
    """
    A function might call other functions decorated with tf_run_allowed.
    This decorator is used to avoid that such functions are evaluated.
    """
    @wraps(f)
    def wrapper(*args, **kwargs):
        # use this context to keep track of the decorated functions calls (recursive depth level).
        # this way, deeper execution level will not be evaluated.
        # finally, in first execution level we do not evaluate the result neither
        with runner_scope():
            # now execute the function and return the result
            return f(*args, **kwargs)

    return wrapper


def set_tf_run(enable):
    # this function is used to modify the default state of tf run (eval tensors or not)
    global __tf_run_default
    __tf_run_default = enable


def try_run(obj):
    try:
        ev_obj = util.get_session().run(obj)
        return ev_obj
    except (RuntimeError, TypeError, ValueError):
        # cannot evaluate the result, return the obj
        return obj



# required pacakges
import inferpy as inf
import numpy as np
import tensorflow as tf


@inf.probmodel
def log_reg(d):
    w0 = inf.Normal(0., 1, name="w0")
    w = inf.Normal(0., 1, batch_shape=[d,1], name="w")

    with inf.datamodel():
        x = inf.Normal(0., 2., batch_shape=d, name="x")
        y = inf.Bernoulli(logits = w0 + x @ w, name="y")




@inf.probmodel
def qmodel(d):
    qw0_loc = inf.Parameter(0., name="qw0_loc")
    qw0_scale = tf.math.softplus(inf.Parameter(1., name="qw0_scale"))
    qw0 = inf.Normal(qw0_loc, qw0_scale, name="w0")

    qw_loc = inf.Parameter(tf.zeros([d,1]), name="qw_loc")
    qw_scale = tf.math.softplus(inf.Parameter(tf.ones([d,1]), name="qw_scale"))
    qw = inf.Normal(qw_loc, qw_scale, name="w")


# create an instance of the model
m = log_reg(d=2)

### create toy data
N = 1000
data = m.sample(size = N, data={"w0":0, "w":[[2],[1]]})
x_train = data["x"]
y_train = data["y"]



VI = inf.inference.VI(qmodel(2), epochs=10000)
m.fit({"x": x_train, "y":y_train}, VI)

sess = inf.get_session()
print(m.posterior["w"].sample())
print(sess.run(m.posterior["w"].loc))



# required pacakges
import inferpy as inf
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt

from inferpy.datasets import mnist

tf.reset_default_graph()
tf.set_random_seed(1234)

# number of components
k = 2
# dimensionality of the data
d = 28*28
# number of observations (dataset size)
N = 1000
# digits considered
DIG = [0,1,2]

learning_rate = 0.01


# definition of a generic model
@inf.probmodel
def pca(k,d):
    w = inf.Normal(loc=tf.zeros([k,d]),
                   scale=1, name="w")               # shape = [k,d]

    w0 = inf.Normal(loc=tf.zeros([d]),
                    scale=1, name="w0")  # shape = [d]

    with inf.datamodel():

        z = inf.Normal(tf.zeros([k]),1, name="z")       # shape = [N,k]
        x = inf.Normal( z @ w + w0, 1, name="x")         # shape = [N,d]



@inf.probmodel
def qmodel(k,d):
    qw_loc = inf.Parameter(tf.zeros([k,d]), name="qw_loc")
    qw_scale = tf.math.softplus(inf.Parameter(tf.ones([k,d]), name="qw_scale"))
    qw = inf.Normal(qw_loc, qw_scale, name="w")

    qw0_loc = inf.Parameter(tf.ones([d]), name="qw0_loc")
    qw0_scale = tf.math.softplus(inf.Parameter(tf.ones([d]), name="qw0_scale"))
    qw0 = inf.Normal(qw0_loc, qw0_scale, name="w0")


    with inf.datamodel():


        qz_loc = inf.Parameter(np.zeros([k]), name="qz_loc")
        qz_scale = tf.math.softplus(inf.Parameter(tf.ones([k]), name="qz_scale"))
        qz = inf.Normal(qz_loc, qz_scale, name="z")



# create an instance of the P model and the Q model
m = pca(k,d)
q = qmodel(k,d)

# load the data
(x_train, y_train), _ = mnist.load_data(num_instances=N, digits=DIG)

optimizer = tf.train.AdamOptimizer(learning_rate)
VI = inf.inference.VI(q, optimizer=optimizer, epochs=2000)

m.fit({"x": x_train}, VI)



# Plot the evolution of the loss

L = VI.losses
plt.plot(range(len(L)), L)

plt.xlabel('epochs')
plt.ylabel('Loss')
plt.title('Loss evolution')
plt.grid(True)
plt.show()


#extract the hidden encoding
sess = inf.get_session()
post = {v:sess.run(m.posterior[v].loc) for v in ["z", "w", "w0"] }


m.posterior["z"].copy()

# plot
markers = ["x", "+", "o"]
colors = [plt.get_cmap("gist_rainbow")(0.05),
          plt.get_cmap("gnuplot2")(0.08),
          plt.get_cmap("gist_rainbow")(0.33)]
transp = [0.9, 0.9, 0.5]

fig = plt.figure()

for c in range(0, len(DIG)):
    col = colors[c]
    plt.scatter(post["z"][y_train == DIG[c], 0], post["z"][y_train == DIG[c], 1], color=col,
                label=DIG[c], marker=markers[c], alpha=transp[c], s=60)
    plt.legend()

plt.show()



with inf.contextmanager.observe(m.posterior, {"z": post["z"]}):
    x_gen = m._last_expanded_vars['x'].sample()

mnist.plot_digits(x_gen, grid=[10,5])






import numpy as np
import inferpy as inf
N=1000
x_train = np.concatenate([inf.Normal([0.0,0.0], scale=1.).sample(int(N/2)), inf.Normal([10.0,10.0], scale=1.).sample(int(N/2))])
##########


# required pacakges
import inferpy as inf
import tensorflow as tf


# number of components
k = 1
# size of the hidden layer in the NN
d0 = 100
# dimensionality of the data
dx = 2
# number of observations (dataset size)
N = 1000


@inf.probmodel
def nlpca(k, d0, dx, decoder):

    with inf.datamodel():
        z = inf.Normal(tf.ones([k])*0.5, 1., name="z")    # shape = [N,k]
        output = decoder(z,d0,dx)
        x_loc = output[:,:dx]
        x_scale = tf.nn.softmax(output[:,dx:])
        x = inf.Normal(x_loc, x_scale, name="x")   # shape = [N,d]


def decoder(z,d0,dx):
    h0 = tf.layers.dense(z, d0, tf.nn.relu)
    return tf.layers.dense(h0, 2 * dx)


# Q-model  approximating P

@inf.probmodel
def qmodel(k):
    with inf.datamodel():
        qz_loc = inf.Parameter(tf.ones([k])*0.5, name="qz_loc")
        qz_scale = tf.math.softplus(inf.Parameter(tf.ones([k]),name="qz_scale"))

        qz = inf.Normal(qz_loc, qz_scale, name="z")


# create an instance of the model
m = nlpca(k,d0,dx, decoder)

# set the inference algorithm
VI = inf.inference.VI(qmodel(k), epochs=5000)

# learn the parameters
m.fit({"x": x_train}, VI)

#extract the hidden representation
hidden_encoding = m.posterior("z")
print(hidden_encoding.sample())

#63




def decoder_keras(z,d0,dx):
    h0 = tf.keras.layers.Dense(d0, activation=tf.nn.relu, name="encoder_h0")
    h1 = tf.keras.layers.Dense(2*dx, name="encoder_h1")
    return h1(h0(z))


# create an instance of the model
m = nlpca(k,d0,dx, decoder_keras)




#80


import numpy as np
import inferpy as inf
N=1000
x_train = np.concatenate([inf.Normal([0.0,0.0], scale=1.).sample(int(N/2)), inf.Normal([10.0,10.0], scale=1.).sample(int(N/2))])
x_test = np.concatenate([inf.Normal([0.0,0.0], scale=1.).sample(int(N/2)), inf.Normal([10.0,10.0], scale=1.).sample(int(N/2))])
##########

import inferpy as inf
import numpy as np
import tensorflow as tf

# definition of a generic model
@inf.probmodel
def pca(k,d):
    beta = inf.Normal(loc=tf.zeros([k,d]),
                   scale=1, name="beta")               # shape = [k,d]

    with inf.datamodel():
        z = inf.Normal(tf.ones([k]),1, name="z")       # shape = [N,k]
        x = inf.Normal(z @ beta , 1, name="x")         # shape = [N,d]


# create an instance of the model
m = pca(k=1,d=2)

@inf.probmodel
def qmodel(k,d):
    qbeta_loc = inf.Parameter(tf.zeros([k,d]), name="qbeta_loc")
    qbeta_scale = tf.math.softplus(inf.Parameter(tf.ones([k,d]),
                                                 name="qbeta_scale"))

    qbeta = inf.Normal(qbeta_loc, qbeta_scale, name="beta")

    with inf.datamodel():
        qz_loc = inf.Parameter(np.ones([k]), name="qz_loc")
        qz_scale = tf.math.softplus(inf.Parameter(tf.ones([k]),
                                                  name="qz_scale"))

        qz = inf.Normal(qz_loc, qz_scale, name="z")

# set the inference algorithm
VI = inf.inference.VI(qmodel(k=1,d=2), epochs=2000)

# learn the parameters
m.fit({"x": x_train}, VI)

# extract the hidden encoding
hidden_encoding = m.posterior("z").parameters()["loc"]

# project x_test into the reduced space (encode)
m.posterior("z", data={"x": x_test}).sample(5)

# sample from the posterior predictive (i.e., simulate values for x given the learnt hidden)
m.posterior_predictive("x").sample(5)

# decode values from the hidden representation
m.posterior_predictive("x", data={"z": [2]}).sample(5)




import inferpy as inf
import numpy as np

# definition of a generic model
@inf.probmodel
def pca(k,d):
    w = inf.Normal(loc=np.zeros([k,d]), scale=1, name="w")      # shape = [k,d]
    with inf.datamodel():
        z = inf.Normal(np.ones(k),1, name="z")                # shape = [N,k]
        x = inf.Normal(z @ w , 1, name="x")                     # shape = [N,d]


# create an instance of the model
m = pca(k=1,d=2)

# create another instance of the model
m2 = pca(k=3, d=10)


# get a sample from the pior distirbution
s = m.sample()



# compute the log probability
m.log_prob(s)


m.sample(size=5)

"""
>>> m.sample(size=5)
{'w': array([[ 1.4588978 , -0.78119284]], dtype=float32), 
 'z': array([[-0.30752778],
       [-0.17902303],
       [ 0.37584424],
       [ 2.1763606 ],
       [ 1.4520675 ]], dtype=float32), 
 'x': array([[-0.7112326 ,  1.3398352 ],
       [ 0.2738361 ,  0.0637802 ],
       [-1.7577622 , -0.26795918],
       [ 3.2689626 , -2.3581226 ],
       [ 2.539008  , -0.13005853]], dtype=float32)}
"""


import tensorflow as tf

from inferpy import util
from inferpy import contextmanager
from .vi import VI


class SVI(VI):
    def __init__(self, *args, batch_size=100, **kwargs):
        """Creates a new Stochastic Variational Inference object.

            Args:
                *args: list of arguments used for the super().__init__ function
                *kwargs: dict of arguments used for the super().__init__ function
                batch_size (`int`): The number of epochs to run in the gradient descent process
        """
        # Build the super object
        super().__init__(*args, **kwargs)

        # and save the extra argument batch size
        self.batch_size = batch_size
        self.plate_size = batch_size  # the plate_size matches the batch_size
        self.batches = None
        self.batch_weight = None

    def compile(self, pmodel, data_size):
        # set the used pmodel
        self.pmodel = pmodel
        # compute the batch_weight depending on the data_size and the batch_size
        self.batch_weight = self.batch_size / data_size  # N/M
        # create the train tensor
        self.train_tensor = self._generate_train_tensor(batch_weight=self.batch_weight)

    def update(self, sample_dict):
        # create the input_data tensor
        input_data = self.create_input_data_tensor(sample_dict)

        t = []
        sess = util.get_session()
        for i in range(self.epochs):
            for j in range(self.batches):
                # evaluate the data tensor to get an evaluated one which can be used to observe varoables
                local_input_data = sess.run(input_data)
                with contextmanager.observe(self.expanded_variables["p"], local_input_data):
                    with contextmanager.observe(self.expanded_variables["q"], local_input_data):
                        sess.run(self.train_tensor)

                        t.append(sess.run(self.debug.loss_tensor))
                        if j == 0 and i % 200 == 0:
                            print("\n {} epochs\t {}".format(i, t[-1]), end="", flush=True)
                        if j == 0 and i % 20 == 0:
                            print(".", end="", flush=True)

        # set the protected _losses attribute for the losses property
        self.debug.losses += t

    def create_input_data_tensor(self, sample_dict):
        # NOTE: data_size, batches and batch_weight can be different in each iteration

        # create a tf dataset and an iterator, specifying the batch size
        data_size = util.iterables.get_plate_size(self.pmodel.vars, sample_dict)
        self.batches = int(data_size / self.batch_size)  # M/N

        # ensure that the number of batches is equal or greater than 1
        if self.batches < 1:
            raise ValueError("The size of the data must be equal or greater than the batch size")

        self.batch_weight = self.batch_size / data_size  # N/M

        tfdataset = (
            tf.data.Dataset.from_tensor_slices(sample_dict)
            .shuffle(data_size)  # use the size of the complete dataset for shuffle buffer, so we use a perfect shuffle
            .batch(self.batch_size, drop_remainder=True)  # discard the remainder batch with less elements if exists
            .repeat()
        )
        iterator = tfdataset.make_one_shot_iterator()
        input_data = iterator.get_next()  # each time this tensor is evaluated in a session it contains new data

        return input_data



import inferpy as inf
import tensorflow as tf
import numpy as np


@inf.probmodel
def simple(mu=0):
    # global variables
    theta = inf.Normal(mu, 0.1, name="theta")

    # local variables
    with inf.datamodel():
        x = inf.Normal(theta, 1, name="x")

m = simple()

#17

"""
>>> m = simple()
>>> type(m)
<class 'inferpy.models.prob_model.ProbModel'>
"""

# 25

"""
>>> m.prior().sample()
{'theta': -0.074800275, 'x': array([0.07758344], dtype=float32)}

>>> m.prior().parameters()
{'theta': {'name': 'theta',
  'allow_nan_stats': True,
  'validate_args': False,
  'scale': 0.1,
  'loc': 0},
 'x': {'name': 'x',
  'allow_nan_stats': True,
  'validate_args': False,
  'scale': 1,
  'loc': 0.116854645}}
"""



"""
>>> m.vars["theta"]
<inf.RandomVariable (Normal distribution) named theta/, shape=(), dtype=float32>
"""





#55

"""
>>> m2 = simple(mu=5)
>>> m==m2
False
"""




#66
"""
>>> sess = inf.get_session()
>>> sess.run(m2.vars["x"].loc)
4.849595
"""

# 73

"""
>>> inf.models.random_variable.distributions_all
['Autoregressive', 'BatchReshape', 'Bernoulli', 'Beta', 'BetaWithSoftplusConcentration',
 'Binomial', 'Categorical', 'Cauchy', 'Chi2', 'Chi2WithAbsDf', 'ConditionalTransformedDistribution',
  'Deterministic', 'Dirichlet', 'DirichletMultinomial', 'ExpRelaxedOneHotCategorical', '
  Exponential', 'ExponentialWithSoftplusRate', 'Gamma', 'GammaGamma', 
  'GammaWithSoftplusConcentrationRate', 'Geometric', 'GaussianProcess', 
  'GaussianProcessRegressionModel', 'Gumbel', 'HalfCauchy', 'HalfNormal', 
  'HiddenMarkovModel', 'Horseshoe', 'Independent', 'InverseGamma',
   'InverseGammaWithSoftplusConcentrationRate', 'InverseGaussian', 'Kumaraswamy',
   'LinearGaussianStateSpaceModel', 'Laplace', 'LaplaceWithSoftplusScale', 'LKJ',
  'Logistic', 'LogNormal', 'Mixture', 'MixtureSameFamily', 'Multinomial',
   'MultivariateNormalDiag', 'MultivariateNormalFullCovariance', 'MultivariateNormalLinearOperator',
   'MultivariateNormalTriL', 'MultivariateNormalDiagPlusLowRank', 'MultivariateNormalDiagWithSoftplusScale',
   'MultivariateStudentTLinearOperator', 'NegativeBinomial', 'Normal', 'NormalWithSoftplusScale', 
   'OneHotCategorical', 'Pareto', 'Poisson', 'PoissonLogNormalQuadratureCompound', 'QuantizedDistribution',
   'RelaxedBernoulli', 'RelaxedOneHotCategorical', 'SinhArcsinh', 'StudentT', 'StudentTWithAbsDfSoftplusScale', 
   'StudentTProcess', 'TransformedDistribution', 'Triangular', 'TruncatedNormal', 'Uniform', 'VectorDeterministic',
   'VectorDiffeomixture', 'VectorExponentialDiag', 'VectorLaplaceDiag', 'VectorSinhArcsinhDiag', 'VonMises', 
   'VonMisesFisher', 'Wishart', 'Zipf']
"""



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import tensorflow as tf
from tensorflow.python.client import session as tf_session

from inferpy import contextmanager
from inferpy import util


class Parameter:
    """
    Random Variable parameter which can be optimized by an inference mechanism.
    """
    def __init__(self, initial_value, name=None):
        # By defult, parameter is not expanded
        self.is_datamodel = False

        # the parameter must have a name
        self.name = name if name else util.name.generate('parameter')

        # check if Parameter is created inside a datamodel context or not.
        if contextmanager.data_model.is_active():
            # In this case, the parameter is in datamodel
            self.is_datamodel = True

            # convert parameter to tensor if it is not
            initial_value = tf.cast(initial_value, tf.float32)

            input_varname = initial_value.op.name if contextmanager.randvar_registry.is_building_graph() else name
            # check the sample_shape. If not empty, expand the initial_value
            contextmanager.randvar_registry.update_graph(input_varname)

            sample_shape = contextmanager.data_model.get_sample_shape(input_varname)
            if sample_shape is not ():
                initial_value = \
                    tf.broadcast_to(initial_value, tf.TensorShape(sample_shape).concatenate(initial_value.shape))

        # Build the tf variable
        self.var = tf.Variable(initial_value, name=self.name)
        util.session.get_session().run(tf.variables_initializer([self.var]))

        # register the variable, which is used to detect dependencies
        contextmanager.randvar_registry.register_parameter(self)
        contextmanager.randvar_registry.update_graph()


def _tensor_conversion_function(p, dtype=None, name=None, as_ref=False):
    """
        Function that converts the inferpy variable into a Tensor.
        This will enable the use of enable tf.convert_to_tensor(rv)

        If the variable needs to be broadcast_to, do it right now
    """
    return tf.convert_to_tensor(p.var)


# register the conversion function into a tensor
tf.register_tensor_conversion_function(  # enable tf.convert_to_tensor
    Parameter, _tensor_conversion_function)


def _session_run_conversion_fetch_function(p):
    """
        This will enable run and operations with other tensors
    """
    return ([tf.convert_to_tensor(p)], lambda val: val[0])


tf_session.register_session_run_conversion_functions(  # enable sess.run, eval
    Parameter,
    _session_run_conversion_fetch_function)



import inferpy as inf
import tensorflow as tf
import numpy as np


x = inf.Normal(loc = 0, scale = 1)

x.var
x.distribution

"""
>>> x = inf.Normal(loc = 0, scale = 1)

>>> x.var
<ed.RandomVariable 'randvar_0/' shape=() dtype=float32>

>>> x.distribution
<tfp.distributions.Normal 'randvar_0/' batch_shape=() event_shape=() dtype=float32>

"""




"""
>>> x.value
<tf.Tensor 'randvar_0/sample/Reshape:0' shape=() dtype=float32>

>>> x.sample()
-0.05060442

>>> x.loc
<tf.Tensor 'randvar_0/Identity:0' shape=() dtype=float32>
"""


"""
>>> x.sample(tf_run=False)
<tf.Tensor 'randvar_0/sample/Reshape:0' shape=() dtype=float32>
"""


### 41





# batch shape

x = inf.Normal(0,1, batch_shape=[3,2])                    # x.shape = [3,2]

x = inf.Normal(loc = [[0.,0.],[0.,0.],[0.,0.]], scale=1)  # x.shape = [3,2]

x = inf.Normal(loc = np.zeros([3,2]), scale=1)            # x.shape = [3,2]

x = inf.Normal(loc = 0, scale=tf.ones([3,2]))             # x.shape = [3,2]





# sample shape


x = inf.Normal(tf.ones([3,2]), 0, sample_shape=100)     # x.sample = [100,3,2]

with inf.datamodel(100):
    x = inf.Normal(tf.ones([3, 2]), 0)                  # x.sample = [100,3,2]



# event shape

x = inf.MultivariateNormalDiag(loc=[1., -1], scale_diag=[1, 2.])



### 63

"""

>>> x.event_shape
TensorShape([Dimension(2)])

>>> x.batch_shape
TensorShape([])

>>> x.sample_shape
TensorShape([])



"""






###  83

with inf.datamodel(size=10):
    x = inf.models.Normal(loc=tf.zeros(5), scale=1.)       # x.shape = [10,5]


y = x[7,4]                                              # y.shape = []

y2 = x[7]                                               # y2.shape = [5]

y3 = x[7,:]                                             # y2.shape = [5]

y4 = x[:,4]                                             # y4.shape = [10]



assert(y.shape==[])
assert(y2.shape==[5])
assert(y3.shape==[5])
assert(y4.shape==[10])


i = inf.Categorical(logits= tf.zeros(3))        # shape = []
mu = inf.Normal([5,1,-2], 0.)                   # shape = [3]
x = inf.Normal(mu[i], scale=1.)                 # shape = []



### 106



@inf.probmodel
def simple(mu=0):
    # global variables
    theta = inf.Normal(mu, 0.1, name="theta")

    # local variables
    with inf.datamodel():
        x = inf.Normal(theta, 1, name="x")







m = simple()

"""
>>> m.sample()
{'theta': -0.074800275, 'x': array([0.07758344], dtype=float32)}
"""



"""
>>> m.vars["theta"]
<inf.RandomVariable (Normal distribution) named theta/, shape=(), dtype=float32>
"""





m2 = simple(mu=5)

"""
>>> sess = tf.session()
>>> sess.run(m2.vars["x"].loc)
4.849595
"""







import numpy as np
import functools

from inferpy import contextmanager
from inferpy import util


def flatten_result(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        simplify_result = kwargs.pop('simplify_result', True)
        result = f(*args, **kwargs)
        if simplify_result and len(result) == 1:
            return result[list(result.keys())[0]]
        else:
            return result
    return wrapper


class Query:
    def __init__(self, variables, target_names=None, data={}):
        # if provided a single name, create a list with only one item
        if isinstance(target_names, str):
            target_names = [target_names]

        # raise an error if target_names is not None and contains variable names not in variables
        if target_names and any((name not in variables for name in target_names)):
            raise ValueError("Target names must correspond to variable names")

        self.target_variables = variables if not target_names else \
            {k: v for k, v in variables.items() if k in target_names}

        self.observed_variables = variables
        self.data = data

    @flatten_result
    @util.tf_run_ignored
    def log_prob(self):
        """ Computes the log probabilities of a (set of) sample(s)"""
        with contextmanager.observe(self.observed_variables, self.data):
            result = util.runtime.try_run({k: v.log_prob(v.value) for k, v in self.target_variables.items()})

        return result

    def sum_log_prob(self):
        """ Computes the sum of the log probabilities (evaluated) of a (set of) sample(s)"""
        # The decorator is not needed here because this function returns a single value
        return np.sum([np.mean(lp) for lp in self.log_prob(simplify_result=False).values()])

    @flatten_result
    @util.tf_run_ignored
    def sample(self, size=1):
        """ Generates a sample for eache variable in the model """
        with contextmanager.observe(self.observed_variables, self.data):
            # each iteration for `size` run the dict in the session, so if there are dependencies among random variables
            # they are computed in the same graph operations, and reflected in the results
            samples = util.runtime.try_run([{k: (v.sample(v.sample_shape) if v.sample_shape else v.sample())
                                             for k, v in self.target_variables.items()}
                                            for _ in range(size)])

        if size == 1:
            result = samples[0]
        else:
            # compact all samples in one single dict
            result = {k: np.concatenate([sample[k] for sample in samples]) for k in self.target_variables.keys()}

        return result

    @flatten_result
    @util.tf_run_ignored
    def parameters(self, names=None):
        """ Return the parameters of the Random Variables of the model.
        If `names` is None, then return all the parameters of all the Random Variables.
        If `names` is a list, then return the parameters specified in the list (if exists) for all the Random Variables.
        If `names` is a dict, then return all the parameters specified (value) for each Random Variable (key).

        NOTE: If tf_run=True, but any of the returned parameters is not a Tensor *and therefore cannot be evaluated)
            this returns a not evaluated dict (because the evaluation will raise an Exception)

        Args:
            names: A list, a dict or None. Specify the parameters for the Random Variables to be obtained.

        Returns:
            A dict, where the keys are the names of the Random Variables and the values a dict of parameters (name-value)
        """
        # argument type checking
        if not(names is None or isinstance(names, (list, dict))):
            raise TypeError("The argument 'names' must be None, a list or a dict, not {}.".format(type(names)))
        # now we can assume that names is None, a list or a dict

        # function to filter the parameters for each Random Variable
        def filter_parameters(varname, parameters):
            parameter_names = list(parameters.keys())
            if names is None:
                # use all the parameters
                selected_parameters = parameter_names
            else:
                # filter by names; if is a dict and key not in, use all the parameters
                selected_parameters = set(names if isinstance(names, list) else names.get(varname, parameters))

            return {k: util.runtime.try_run(v) for k, v in parameters.items() if k in selected_parameters}

        with contextmanager.observe(self.observed_variables, self.data):
            result = {k: filter_parameters(k, v.parameters)
                      for k, v in self.target_variables.items()}

        return result



import tensorflow as tf
import inspect
import itertools
from tensorflow_probability.python import edward2 as ed

from . import loss_functions
import inferpy as inf
from inferpy import util
from inferpy import contextmanager

from ..inference import Inference


class VI(Inference):
    def __init__(self, qmodel, loss='ELBO', optimizer='AdamOptimizer', epochs=1000):
        """Creates a new Variational Inference object.

            Args:
                qmodel (`inferpy.ProbModel`): The q model
                loss (`str` or `function`): A function that computes the loss tensor from the expanded variables
                    of p and q models, or a `str` that refears to a function with that name in the package
                    `inferpy.inference.variational.loss_function`
                optimizer (`str` or `tf.train.Optimizer`): An optimizer object from `tf.train` optimizers, or a string
                    that refers to the name of an optimizer in such module or package
                epochs (`int`): The number of epochs to run in the gradient descent process
        """

        # store the qmodel in self.qmodel. Can be a callable with no parameters which returns the qmodel
        if callable(qmodel):
            if len(inspect.signature(qmodel).parameters) > 0:
                raise ValueError("input qmodel can only be a callable object if this does not has any input parameter")
            self.qmodel = qmodel()
        else:
            self.qmodel = qmodel

        # store the loss function in self.loss_fn
        # if it is a string, build the object automatically from loss_functions package
        if isinstance(loss, str):
            self.loss_fn = getattr(loss_functions, loss)
        else:
            self.loss_fn = loss

        self.epochs = epochs

        # store the optimizer function in self.optimizer
        # if it is a string, build a new optimizer from tf.train (default parametrization)
        if isinstance(optimizer, str):
            self.optimizer = getattr(tf.train, optimizer)()
        else:
            self.optimizer = optimizer

        # pmodel not established yet
        self.pmodel = None
        # The size of the plate when expand the models
        self.plate_size = None
        # The tensor to optimize the tf.Variables
        self.train_tensor = None

        # expanded variables and parameters
        self.expanded_variables = {"p": None, "q": None}
        self.expanded_parameters = {"p": None, "q": None}

        # list for storing the loss evolution
        class Debug:
            pass
        self.debug = Debug()

        self.debug.losses = []

    def compile(self, pmodel, data_size):
        # set the used pmodel
        self.pmodel = pmodel
        # and the plate size, which matches the data size
        self.plate_size = data_size
        # create the train tensor
        self.train_tensor = self._generate_train_tensor(plate_size=self.plate_size)

    def update(self, sample_dict):
        # ensure that the size of the data matches with the self.plate_size
        data_size = util.iterables.get_plate_size(self.pmodel.vars, sample_dict)
        if data_size != self.plate_size:
            raise ValueError("The size of the data must be equal to the plate size: {}".format(self.plate_size))

        t = []
        sess = util.get_session()
        with contextmanager.observe(self.expanded_variables["p"], sample_dict):
            with contextmanager.observe(self.expanded_variables["q"], sample_dict):
                for i in range(self.epochs):
                    sess.run(self.train_tensor)

                    t.append(sess.run(self.debug.loss_tensor))
                    if i % 200 == 0:
                        print("\n {} epochs\t {}".format(i, t[-1]), end="", flush=True)
                    if i % 10 == 0:
                        print(".", end="", flush=True)

        # set the protected _losses attribute for the losses property
        self.debug.losses += t

    @property
    def losses(self):
        return self.debug.losses

    def sample(self, size=1, data={}):
        raise NotImplementedError

    def log_prob(self, data):
        raise NotImplementedError

    def parameters(names=None):
        raise NotImplementedError

    def _generate_train_tensor(self, **kwargs):
        """ This function expand the p and q models. Then, it uses the  loss function to create the loss tensor
            and store it into the debug object as a new attribute.
            Then, uses the optimizer to create the train tensor used in the gradient descent iterative process.
            It store the expanded random variables and parameters from the p and q models in self.expanded_variables
            and self.expanded_parameters dicts.

            Returns:
                The `tf.Tensor` train tensor used in the gradient descent iterative process.

        """
        # expand the p and q models
        # expand de qmodel
        qvars, qparams = self.qmodel.expand_model(self.plate_size)

        # expand de pmodel, using the intercept.set_values function, to include the sample_dict and the expanded qvars
        with ed.interception(util.interceptor.set_values(**qvars)):
            pvars, pparams = self.pmodel.expand_model(self.plate_size)

        # create the loss tensor and trainable tensor for the gradient descent process
        loss_tensor = self.loss_fn(pvars, qvars, **kwargs)
        train = self.optimizer.minimize(loss_tensor)

        # save the expanded variables and parameters
        self.expanded_variables = {
            "p": pvars,
            "q": qvars
        }
        self.expanded_parameters = {
            "p": pparams,
            "q": qparams
        }
        # save the loss tensor for debug purposes
        self.debug.loss_tensor = loss_tensor

        # Initialize all variables which are not in the probmodel p, because they have been initialized before
        model_variables = [v for v in itertools.chain(
            self.pmodel.params.values(),  # do not re-initialize prior p model parameters
            pparams.values(),  # do not re-initialize posterior p model parameters
            self.qmodel.params.values(),  # do not re-initialize prior q model parameters
            qparams.values(),  # do not re-initialize posterior q model parameters
           )]
        inf.get_session().run(
            tf.variables_initializer([
                v for v in tf.global_variables() if v not in model_variables and not v.name.startswith("inferpy-")
                ]))

        return train



import numpy as np
import pytest

import inferpy as inf
from inferpy import util
from tests import no_raised_exc


def test_build_model():
    sample_size = 100
    batch_shape = (2, 3)

    @inf.probmodel
    def model():
        p = inf.Parameter(np.zeros(batch_shape, dtype=np.float32), name='p')
        with inf.datamodel():
            x = inf.Normal(p, 1., name='x')
            inf.Normal(x, 1., name='y')

    m = model()
    # builder is a function
    assert callable(m.builder)
    # vars is an OrderedDict with the random variables
    assert 'x' in m.vars
    assert 'y' in m.vars
    # params is an OrderedDict with the parameters
    assert 'p' in m.params

    # if fit for datamodel not used, then the shape of random vars is (1, batch_shape)
    assert tuple(m.vars['x'].shape.as_list()) == (1, ) + batch_shape
    assert tuple(m.vars['y'].shape.as_list()) == (1, ) + batch_shape
    # if expanded datamodel, then the shape of random vars is (sample_shape, batch_shape)
    expanded_vars, _ = m.expand_model(sample_size)
    assert tuple(expanded_vars['x'].shape.as_list()) == (sample_size, ) + batch_shape
    assert tuple(expanded_vars['y'].shape.as_list()) == (sample_size, ) + batch_shape

    # assert variables and parameters are in graph
    assert 'x' in m.graph
    assert 'y' in m.graph
    assert 'p' in m.graph


@pytest.mark.parametrize("data, expected_flow, expected_result", [
    # empty dict
    (
        dict(),
        no_raised_exc(),
        1
    ),
    # one field which does not exist
    (
        dict(othername=np.ones(100)),
        no_raised_exc(),
        1
    ),
    # one variable data size
    (
        dict(x=np.ones(100)),
        no_raised_exc(),
        100
    ),
    # two variable data with the same size
    (
        dict(x=np.ones(100), y=np.ones(100)),
        no_raised_exc(),
        100
    ),
    # two variable data with different size
    (
        dict(x=np.ones(100), y=np.ones(150)),
        pytest.raises(ValueError),
        None
    ),
])
def test_probmodel_get_plate_size(data, expected_flow, expected_result):
    @inf.probmodel
    def model():
        p = inf.Parameter(0., name='p')
        with inf.datamodel():
            x = inf.Normal(p, 1., name='x')
            inf.Normal(x, 1., name='y')

    with expected_flow:
        m = model()

        plate_size = util.iterables.get_plate_size(m.vars, data)
        assert expected_result == plate_size


def test_sample():
    @inf.probmodel
    def model():
        p = inf.Parameter(0., name='p')
        with inf.datamodel():
            x = inf.Normal(p, 1., name='x')
            inf.Normal(x, 1., name='y')

    N = 100
    m = model()

    sample_dict = m.prior().sample(N)

    varnames = list(sample_dict.keys())
    assert len(varnames) == 2
    assert 'x' in varnames
    assert 'y' in varnames

    assert len(sample_dict['x']) == N
    assert len(sample_dict['y']) == N


def test_sample_intercept():
    @inf.probmodel
    def model():
        p = inf.Parameter(0., name='p')
        with inf.datamodel():
            x = inf.Normal(p, 1., name='x')
            inf.Normal(x, 1., name='y')

    N = 10
    data_y = 1.0
    m = model()

    sample_dict = m.prior(data={'y': data_y}).sample(N)

    varnames = list(sample_dict.keys())
    assert len(varnames) == 2
    assert 'x' in varnames
    assert 'y' in varnames

    assert len(sample_dict['x']) == N
    assert len(sample_dict['y']) == N


def test_log_prob():
    @inf.probmodel
    def model():
        p = inf.Parameter(0., name='p')
        with inf.datamodel():
            x = inf.Normal(p, 1., name='x')
            inf.Normal(x, 1., name='y')

    m = model()

    data = m.prior(['x', 'y']).sample()
    print(data)
    logprob_dict = m.prior(['x', 'y']).log_prob()
    varnames = list(logprob_dict.keys())
    assert len(varnames) == 2
    assert 'x' in varnames
    assert 'y' in varnames

    assert logprob_dict['x'] <= 0.0
    assert logprob_dict['y'] <= 0.0

    # assert that the result of sum_log_prob is a single float32 number
    assert isinstance(m.prior(data=data).sum_log_prob(), np.float32)



import tensorflow as tf
from tensorflow_probability import edward2 as ed


# this function is used to intercept the value property of edward2 random variables
def set_values(**model_kwargs):
    """Creates a value-setting interceptor. Usable as a parameter of the ed2.interceptor.

        :model_kwargs: The name of each argument must be the name of a random variable to intercept,
            and the value is the element which intercepts the value of the random variable.

        :returns: The random variable with the intercepted value
    """

    def interceptor(f, *args, **kwargs):
        """Sets random variable values to its aligned value."""
        name = kwargs.get("name")

        # if name in model_kwargs, include the value as a new argument using model_kwargs[name]
        if name in model_kwargs:
            kwargs["value"] = model_kwargs[name]

        randvar = ed.interceptable(f)(*args, **kwargs)

        # if the value used to intercept is an inferpy RandomVariable, assign to the object
        # the tf.Variables to observe the Random Variable as well
        if name in model_kwargs and \
                hasattr(model_kwargs[name], 'is_observed') and hasattr(model_kwargs[name], 'observed_value'):
            randvar.is_observed = model_kwargs[name].is_observed
            randvar.observed_value = model_kwargs[name].observed_value

        # finally return the built random variable
        return randvar

    return interceptor


# this function is used to intercept the value property of edward2 random variables
# dependent on a tf variable var_condition: if true use var_value, otherwise use the variable value
def set_values_condition(var_condition, var_value):
    """Creates a value-setting interceptor. Usable as a parameter of the ed2.interceptor.

        :var_condition (`tf.Variable`): The boolean tf.Variable, used to intercept the value property with
            `value_var` or the variable value property itself
        :var_value (`tf.Variable`): The tf.Variable used to intercept the value property when `var_condition` is True

        :returns: The random variable with the intercepted value
    """

    def interceptor(f, *args, **kwargs):
        """Sets random variable values to its aligned value."""

        # need to create the variable to obtain its value, and use it in the fn_false condition
        _value = ed.interceptable(f)(*args, **kwargs).value

        conditional_value = tf.cond(
            var_condition,
            lambda: var_value,
            lambda: _value
        )

        # need to broadcast to fix the shape of the tensor (which always be _value.shape)
        return ed.interceptable(f)(*args, value=tf.broadcast_to(conditional_value, _value.shape), **kwargs)

    return interceptor



# required pacakges
import inferpy as inf
import numpy as np
import tensorflow as tf

import matplotlib.pyplot as plt
from inferpy.datasets import mnist



# number of components
k = 2
# size of the hidden layer in the NN
d0 = 100
# dimensionality of the data
dx = 28*28
# number of observations (dataset size)
N = 1000

# digits considered
DIG = [0,1,2]

scale_epsilon = 0.01

@inf.probmodel
def nlpca(k, d0, dx, decoder):
    with inf.datamodel():
        z = inf.Normal(tf.ones([k]) * 0.1, 1., name="z")  # shape = [N,k]
        output = decoder(z, d0, dx)
        x_loc = output[:, :dx]
        x_scale = tf.nn.softmax(output[:, dx:]) + scale_epsilon
        x = inf.Normal(x_loc, x_scale, name="x")  # shape = [N,d]


def decoder(z, d0, dx):  # k -> d0 -> 2*dx
    h0 = tf.layers.dense(z, d0, tf.nn.relu)
    return tf.layers.dense(h0, 2 * dx)


@inf.probmodel
def qmodel(k):
    with inf.datamodel():
        qz_loc = inf.Parameter(np.ones([k])*0.1, name="qz_loc")
        qz_scale = tf.math.softmax(inf.Parameter(tf.ones([k]), name="qz_scale"))
        qz = inf.Normal(qz_loc, qz_scale, name="z")




m = nlpca(k, d0, dx, decoder)
q = qmodel(k)

# set the inference algorithm
VI = inf.inference.VI(q, epochs=5000)

(x_train, y_train), _ = mnist.load_data(num_instances=N, digits=DIG)

# plot the digits
mnist.plot_digits(x_train)

# learn the parameters
m.fit({"x": x_train}, VI)


# Plot the evolution of the loss

L = VI.losses
plt.plot(range(len(L)), L)

plt.xlabel('epochs')
plt.ylabel('Loss')
plt.title('Loss evolution')
plt.grid(True)
plt.show()


#extract the hidden encoding
sess = inf.get_session()
post = {"z":sess.run(m.posterior["z"].loc)}


# plot
markers = ["x", "+", "o"]
colors = [plt.get_cmap("gist_rainbow")(0.05),
          plt.get_cmap("gnuplot2")(0.08),
          plt.get_cmap("gist_rainbow")(0.33)]
transp = [0.9, 0.9, 0.5]

fig = plt.figure()

for c in range(0, len(DIG)):
    col = colors[c]
    plt.scatter(post["z"][y_train == DIG[c], 0], post["z"][y_train == DIG[c], 1], color=col,
                label=DIG[c], marker=markers[c], alpha=transp[c], s=60)
    plt.legend()

plt.show()



with inf.contextmanager.observe(m.posterior, {"z": post["z"]}):
    x_gen = m._last_expanded_vars['x'].sample()

mnist.plot_digits(x_gen, grid=[10,5])




import pytest
import numpy as np

from inferpy.util.iterables import get_shape


@pytest.mark.parametrize('x, expected', [
    # an integer
    (1, ()),
    # a float
    (1.0, ()),
    # a string
    ("foo", ()),
    # a numpy number
    (np.int(1), ()),
    # a numpy array single dimension
    (np.ones(3), (3)),
    # a numpy array two dimension
    (np.ones((2, 3)), (2, 3)),
    # a numpy array three dimension
    (np.ones((2, 3, 4)), (2, 3, 4)),
    # a dict
    (dict(x=1), ()),
    # a dict with a list
    (dict(x=[1, 2, 3]), ()),
    # an empty list
    ([], ()),
    # a list
    ([1, 2, 3], (3)),
    # a list in a list
    ([[1, 2], [3, 4], [5, 6]], (3, 2)),
])
def test_iterables(x, expected):
    get_shape(x) == expected


def test_iterables_exception():
    x = [[1, 2], [3]]  # different number of elements per index in 0-dimension
    with pytest.raises(ValueError):
        get_shape(x)



import pytest
from tensorflow_probability.python import edward2 as ed

from inferpy.util import tf_graph


# elements in the same second lists are dependent
@pytest.mark.parametrize("dependencies1, dependencies2", [
    # two independent variables
    (['x'], ['y']),
    # two sets of independent variables, and each group of three dependent variables (chain)
    (['x', 'y', 'z'], ['u', 'v', 'w'])
])
def test_dependencies(dependencies1, dependencies2):
    # register and check dependencies between group of dependent vars
    for dependent_vars in (dependencies1, dependencies2):
        last = ed.Normal(0, 1, name=dependent_vars[0])
        for i in range(1, len(dependent_vars)):
            parentname = dependent_vars[i - 1]
            name = dependent_vars[i]
            x = ed.Normal(last, 1, name=name)
            g = tf_graph.get_graph(set([parentname, name]))
            assert parentname in g.predecessors(name)
            last = x
    # all variables registered. Now check independencies between independent groups
    g = tf_graph.get_graph(set(dependencies1 + dependencies2))
    for v1 in dependencies1:
        for v2 in dependencies2:
            assert v1 not in g.predecessors(v2)
            assert v2 not in g.predecessors(v1)



"""MNIST handwritten digits dataset."""

from keras.datasets import mnist
import numpy as np
import matplotlib.pyplot as plt


def load_data(vectorize=True, num_instances=None, num_instances_test=None, digits=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]):
    """ Loads the MNIST datase

    :param vectorize: if true, each 2D image is transformed into a 1D vector
    :param num_instances: total number of images loaded
    :param digits: list of integers indicating the digits to be considered
    :return:  Tuple of Numpy arrays: `(x_train, y_train), (x_test, y_test)
    """

    (x_train, y_train), (x_test, y_test) = mnist.load_data()
    (x_train, y_train) = _preprocess_data(x_train, y_train, vectorize, num_instances, digits)
    (x_test, y_test) = _preprocess_data(x_test, y_test, vectorize, num_instances_test, digits)

    return (x_train, y_train), (x_test, y_test)


def _preprocess_data(x_data, y_data, vectorize=True, num_instances=None, digits=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]):

    num_pixels = np.prod(np.shape(x_data)[1:])
    x_data = x_data[np.isin(y_data, digits)]
    y_data = y_data[np.isin(y_data, digits)]

    if num_instances is None:
        num_instances = len(x_data)

    y_data = y_data[:num_instances]
    x_data = x_data[:num_instances]

    if vectorize:
        x_data = np.reshape(x_data, (num_instances, num_pixels))  # serialize the data
        x_data = np.float32(x_data)

    return x_data, y_data


def plot_digits(data, grid=[3, 3]):
    nx, ny = grid
    fig, ax = plt.subplots(nx, ny, figsize=(12, 12))
    fig.tight_layout(pad=0.3, rect=[0, 0, 0.9, 0.9])
    for x, y in [(i, j) for i in list(range(nx)) for j in list(range(ny))]:
        img_i = data[x + y * nx].reshape((28, 28))
        i = (y, x) if nx > 1 else y
        ax[i].imshow(img_i, cmap='gray')
    plt.show()



__version__ = '1.1.0-rc'


from .models import *  # noqa F401, F403
from . import inference  # noqa F401
from .contextmanager import datamodel  # noqa F401
from .util.runtime import set_tf_run  # noqa F401
from .util.session import *  # noqa F401



def get_shape(x):
    """
    Get the shape of an element x. If it is an element with a shape attribute, return it. If it is a list with more than
    one element, compute the shape by checking the len, and the shape of internal elements. In that case, the shape must
    be consistent. Finally, in other case return () as shape.

    :param x: The element to compute its shape
    :raises : class `ValueError`: list shape not consistent
    :returns: A tuple with the shape of `x`
    """
    if isinstance(x, list) and len(x) > 0:
        shapes = [get_shape(subx) for subx in x]
        if any([s != shapes[0] for s in shapes[1:]]):
            raise ValueError('Parameter dimension not consistent: {}'.format(x))
        return (len(x), ) + shapes[0]
    else:
        if hasattr(x, '_shape_tuple'):
            return x._shape_tuple()  # method to return the shape as a tuple
        elif hasattr(x, 'shape'):
            return tuple(x.shape)
        else:
            return ()


def get_plate_size(variables, sample_dict):
        # get the plate size by analyzing the sample_dict input
        # check that all values in dict whose name is a datamodel RV has the same length (will be the plate size)
        plate_shapes = [get_shape(v) for k, v in sample_dict.items()
                        if k in variables and variables[k].is_datamodel]
        plate_sizes = [s[0] if len(s) > 0 else 1 for s in plate_shapes]  # if the shape is (), it is just one element
        if len(plate_sizes) == 0:
            return 1
        else:
            plate_size = plate_sizes[0]

            if any(plate_size != x for x in plate_sizes[1:]):
                raise ValueError('The number of elements for each mapped variable must be the same.')

            return plate_size



"""
Any inference class must implement a run method, which receives a sample_dict object,
and returns a dict of posterior objects (random distributions, list of samples, etc.)
"""

from .variational.vi import VI
from .variational.svi import SVI

__all__ = [
    'SVI',
    'VI'
]



import inferpy as inf
import tensorflow as tf

# number of components
k = 1
# size of the hidden layer in the NN
d0 = 100
# dimensionality of the data
dx = 2
# number of observations (dataset size)
N = 1000


@inf.probmodel
def nlpca(k, d0, dx, decoder):

    with inf.datamodel():
        z = inf.Normal(tf.ones([k])*0.5, 1., name="z")    # shape = [N,k]
        output = decoder(z,d0,dx)
        x_loc = output[:,:dx]
        x_scale = tf.nn.softmax(output[:,dx:])
        x = inf.Normal(x_loc, x_scale, name="x")   # shape = [N,d]


###########  25

def decoder(z,d0,dx):
    h0 = tf.layers.dense(z, d0, tf.nn.relu)
    return tf.layers.dense(h0, 2 * dx)


######### 32

# Q-model  approximating P

@inf.probmodel
def qmodel(k):
    with inf.datamodel():
        qz_loc = inf.Parameter(tf.ones([k])*0.5, name="qz_loc")
        qz_scale = tf.math.softplus(inf.Parameter(tf.ones([k]),name="qz_scale"))

        qz = inf.Normal(qz_loc, qz_scale, name="z")



####### 46

# create an instance of the model
m = nlpca(k,d0,dx, decoder)

# Sample from priors
samples = m.sample()


#### NOT showing  55
import numpy as np
x_train = np.concatenate([inf.Normal([0.0,0.0], scale=1.).sample(int(N/2)), inf.Normal([10.0,10.0], scale=1.).sample(int(N/2))])


###### 60

# set the inference algorithm
VI = inf.inference.VI(qmodel(k), epochs=5000)

# learn the parameters
m.fit({"x": x_train}, VI)


#### 69

#extract the hidden representation
hidden_encoding = m.posterior["z"]
print(hidden_encoding.sample())


####


import contextlib
import numpy as np
from inferpy import util


@contextlib.contextmanager
def observe(variables, data):
    # default session
    sess = util.session.get_session()
    # iterate for all variables both in `variables` and `data`
    for k, v in data.items():
        if k not in variables:
            continue
        # Set the variable to observed. This `tf.Variable` is used by the interceptor `set_values_condition`
        variables[k].is_observed.load(True, session=sess)
        # Now load the value into the `tf.Variable`:
        # if has shape attr:
        if hasattr(v, 'shape'):
            # shape of tf.Variable and value matches
            if v.shape == variables[k].observed_value.shape:
                variables[k].observed_value.load(v, session=sess)
            # shape of tf.Variable and value without the sample_shape (first dim) matches
            # NOTE: this might happend if data comes from sample() and sample_shape == 1
            elif len(v.shape) > 0 and v.shape[0] == 1 and v.shape[1:] == variables[k].observed_value.shape:
                variables[k].observed_value.load(v[0], session=sess)
            # otherwise, just try to do broadcast and load the value
            else:
                # try to broadcast v using numpy (it cannot be a tensor)
                variables[k].observed_value.load(
                    np.broadcast_to(v, variables[k].observed_value.shape.as_list()), session=sess)
        else:
            # try to broadcast v using numpy (it cannot be a tensor)
            variables[k].observed_value.load(
                np.broadcast_to(v, variables[k].observed_value.shape.as_list()), session=sess)
    try:
        yield
    finally:
        # just needs to revert the `is_observed` tf.Variable
        for k, v in data.items():
            if k not in variables:
                continue
            variables[k].is_observed.load(False, session=sess)



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import tensorflow as tf


class Inference:
    """This class implements the functionality of any Inference class.
    """

    def __init__(self):
        raise NotImplementedError

    def compile(self, pmodel, data_size):
        raise NotImplementedError

    def update(self, sample_dict):
        raise NotImplementedError

    def sample(self, size=1, data={}):
        raise NotImplementedError

    def log_prob(self, data):
        raise NotImplementedError

    def sum_log_prob(self, data):
        """ Computes the sum of the log probabilities of a (set of) sample(s)"""
        return tf.reduce_sum([tf.reduce_mean(lp) for lp in self.log_prob(data).values()])

    def parameters(names=None):
        raise NotImplementedError



# -*- coding: utf-8 -*-
#


from setuptools import setup

import os
import re
import sys


if sys.version_info < (3, 4):
    sys.exit('Python < 3.4 is not supported')


# get abs path from this folder name
here = os.path.dirname(os.path.abspath(__file__))

# open __init__.py, where version is specified
with open(os.path.join(here, 'inferpy', '__init__.py')) as f:
    txt = f.read()

# try to read it from source code
try:
    version = re.findall(r"^__version__ = '([^']+)'\r?$",
                         txt, re.M)[0]
except IndexError:
    raise RuntimeError('Unable to determine version.')

# get long description from file in docs folder
with open(os.path.join(here, 'docs/project_description.md')) as f:
    long_description = f.read()


# function to read requirements, and include them as package dependencies
def get_requirements(file):
    # read requirements.txt file and return them as a list of strings
    with open(file) as f:
        req = f.readlines()
    return [r.strip() for r in req]  # clean lines from blank spaces and line breaks


setup(
    name='inferpy',
    version=version,
    description='Probabilistic modeling with Tensorflow made easy',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Andrés R. Masegosa, Rafael Cabañas, Javier Cózar',
    author_email="andresma@ual.es, rcabanas@ual.es, jcozar87@ual.es",
    url='http://inferpy.readthedocs.io',
    download_url='https://github.com/PGMLabSpain/InferPy/archive/{}.tar.gz'.format(version),
    keywords='machine learning statistics probabilistic programming tensorflow edward2',
    license='Apache License 2.0',
    classifiers=['Intended Audience :: Developers',
                 'Intended Audience :: Education',
                 'Intended Audience :: Science/Research',
                 'License :: OSI Approved :: Apache Software License',
                 'Operating System :: POSIX :: Linux',
                 'Operating System :: MacOS :: MacOS X',
                 'Operating System :: Microsoft :: Windows',
                 'Programming Language :: Python :: 3.4'],
    packages=['inferpy'],
    python_requires='>=3.5',
    install_requires=get_requirements('requirements/prod.txt'),
    extras_require={
        'gpu': get_requirements('requirements/gpu.txt'),
        'visualization': get_requirements('requirements/visualization.txt')
    },
    tests_require=get_requirements('requirements/test.txt'),
    include_package_data=True,
)



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import functools
from collections import OrderedDict
from tensorflow_probability import edward2 as ed
import tensorflow as tf
import networkx as nx
from matplotlib import pyplot as plt
import warnings

from inferpy import util
from inferpy import contextmanager
from inferpy.queries import Query
from .random_variable import RandomVariable


def probmodel(builder):
    """
    Decorator to create probabilistic models. The function decorated
    must be a function which declares the Random Variables in the model.
    It is not needed that the function returns such variables (we capture
    them using ed.tape).
    """
    @functools.wraps(builder)
    def wrapper(*args, **kwargs):
        @util.tf_run_ignored
        def fn():
            return builder(*args, **kwargs)
        return ProbModel(
            builder=fn
        )
    return wrapper


class ProbModel:
    """
    Class that implements the probabilistic model functionality.
    It is composed of a graph, capturing the variable relationships, an OrderedDict containing
    the Random Variables/Parameters in order of creation, and the function which declare the
    Random Variables/Parameters.
    """

    def __init__(self, builder):
        # Initialize object attributes
        self.builder = builder
        g_for_nxgraph = tf.Graph()
        # first buid the graph of dependencies
        with g_for_nxgraph.as_default():
            with tf.Session() as sess:
                try:
                    default_sess = util.session.swap_session(sess)
                    self.graph = self._build_graph()
                finally:
                    # sess is closed by context, no need to use set_session (which closes the actual running session)
                    util.session.swap_session(default_sess)

        # Now initialize vars and params for the model (no sample_shape)
        self.vars, self.params = self._build_model()

        # This attribute contains the inference method used. If it is None, the `fit` function has not been used yet
        self.inference_method = None

        self.observed_vars = []  # list of variables that have been observed during the inference

    # all the results of prior, posterior and posterior_predictive are evaluated always, because they depends on
    # tf.Variables, and therefore a tensor cannot be return because the results would depend on the value of that
    # tf.Variables

    def prior(self, target_names=None, data={}):
        return Query(self.vars, target_names, data)

    def posterior(self, target_names=None, data={}):
        if self.inference_method is None:
            raise RuntimeError("posterior cannot be used before using the fit function.")

        # only non-observed variables can be in target_names
        if target_names is None:
            target_names = [name for name in self.vars.keys() if name not in self.observed_vars]
        else:
            if any(var in self.observed_vars for var in target_names):
                raise ValueError("target_names must correspond to not observed variables during the inference: \
                    {}".format([v for v in self.vars.keys() if v not in self.observed_vars]))

        prior_data = self._create_hidden_observations(target_names, data)

        return Query(self.inference_method.expanded_variables["q"], target_names, {**data, **prior_data})

    def posterior_predictive(self, target_names=None, data={}):
        if self.inference_method is None:
            raise RuntimeError("posterior_preductive cannot be used before using the fit function.")

        # only non-observed variables can be in target_names
        if target_names is None:
            target_names = [name for name in self.vars.keys() if name in self.observed_vars]
        else:
            if any(var not in self.observed_vars for var in target_names):
                raise ValueError("target_names must correspond to observed variables during the inference: \
                    {}".format(self.observed_vars))

        prior_data = self._create_hidden_observations(target_names, data)

        return Query(self.inference_method.expanded_variables["p"], target_names, {**data, **prior_data})

    def _create_hidden_observations(self, target_names, data={}):
        # TODO: This code must be implemented independent of the inference method. Right now we are using the p and q
        # expanded variables, which belongs only to variational inference methods. When a different VI is implemented
        # think about a better way to implement this function and access to the correct dict of random variables

        # NOTE: implementation trick. As p model variables are intercepted with q model variables,
        # compute prior observations for local hidden variables which are not targets,
        # expanding a new model using plate_size and then sampling
        hidden_variable_names = [k for k in self.vars.keys() if k not in target_names and k not in data]
        if hidden_variable_names:
            expanded_vars, _ = self.expand_model(self.inference_method.plate_size)
            prior_data = Query(expanded_vars, hidden_variable_names, data).sample(simplify_result=False)
        else:
            prior_data = {}

        return prior_data

    def _build_graph(self):
        with contextmanager.randvar_registry.init():
            self.builder()
            # ed2 RVs created. Relations between them captured in randvar_registry builder as a networkx graph
            nx_graph = contextmanager.randvar_registry.get_graph()

        return nx_graph

    def _build_model(self):
        # get the global variables defined before building the model
        _before_global_variables = tf.global_variables()

        with contextmanager.randvar_registry.init(self.graph):
            # use edward2 model tape to capture RandomVariable declarations
            with ed.tape() as model_tape:
                self.builder()

            # get variables from parameters
            var_parameters = contextmanager.randvar_registry.get_var_parameters()

            # wrap captured edward2 RVs into inferpy RVs
            model_vars = OrderedDict()
            for k, v in model_tape.items():
                registered_rv = contextmanager.randvar_registry.get_variable(k)
                if registered_rv is None:
                    # a ed Random Variable. Create a inferpy Random Variable and assign the var directly.
                    # do not know the args and kwars used to build the ed random variable. Use None.
                    model_vars[k] = RandomVariable(v, name=k, is_datamodel=False, ed_cls=None,
                                                   var_args=None, var_kwargs=None, sample_shape=())
                else:
                    model_vars[k] = registered_rv

        # get the global variables defined after building the model
        _after_global_variables = tf.global_variables()
        # compute the new global variables defined when building the model
        created_vars = [v for v in _after_global_variables if v not in _before_global_variables]
        util.get_session().run(tf.variables_initializer(created_vars))

        return model_vars, var_parameters

    def plot_graph(self):
        nx.draw(self.graph, cmap=plt.get_cmap('jet'), with_labels=True)
        plt.show()

    @util.tf_run_ignored
    def fit(self, sample_dict, inference_method):
        # Parameter checkings
        # sample_dict must be a non empty python dict
        if not isinstance(sample_dict, dict):
            raise TypeError('The `sample_dict` type must be dict.')
        if len(sample_dict) == 0:
            raise ValueError('The number of mapped variables must be at least 1.')

        # if fit was called before, warn that it restarts everything
        if self.inference_method:
            warnings.warn("Fit was called before. This will restart the inference method and \
                re-build the expanded model.")

        # set the inference method
        self.inference_method = inference_method

        # get the plate size
        plate_size = util.iterables.get_plate_size(self.vars, sample_dict)
        # compile the inference method
        inference_method.compile(self, plate_size)
        # and run the update method with the data
        inference_method.update(sample_dict)

        # If it works, set the observed variables
        self.observed_vars = list(sample_dict.keys())

    @util.tf_run_ignored
    def update(self, sample_dict):
        # Check that fit was called first
        if self.inference_method is None:
            raise RuntimeError("The fit method must be called before update")

        # check that the observed_vars are the same
        if set(self.observed_vars) != sample_dict.keys():
            raise ValueError("The data in sample dict must contain only data from observed variables: \
                {}.".format(self.observed_vars))

        # Run the inference method
        self.inference_method.update(sample_dict)

    def expand_model(self, size=1):
        """ Create the expanded model vars using size as plate size and return the OrderedDict """

        with contextmanager.data_model.fit(size=size):
            expanded_vars, expanded_params = self._build_model()

        return expanded_vars, expanded_params



import tensorflow as tf


def ELBO(pvars, qvars, batch_weight=1, **kwargs):
    """ Compute the loss tensor from the expanded variables of p and q models.
        Args:
            pvars (`dict<inferpy.RandomVariable>`): The dict with the expanded p random variables
            qvars (`dict<inferpy.RandomVariable>`): The dict with the expanded q random variables
            batch_weight (`float`): Weight to assign less importance to the energy, used when processing data in batches

        Returns (`tf.Tensor`):
            The generated loss tensor
    """

    # compute energy
    energy = tf.reduce_sum(
        [(batch_weight if p.is_datamodel else 1) * tf.reduce_sum(p.log_prob(p.value))
         for p in pvars.values()])

    # compute entropy
    entropy = - tf.reduce_sum(
        [(batch_weight if q.is_datamodel else 1) * tf.reduce_sum(q.log_prob(q.value))
         for q in qvars.values() if not q.is_datamodel])

    # compute ELBO
    ELBO = energy + entropy

    # This function will be minimized. Return minus ELBO
    return -ELBO



from collections import defaultdict
import networkx as nx
import tensorflow as tf


"""
These set of private functions obtain the dependencies between Random Variables
by analyzing the tensors in Random Varibles and its relations. Finally, the
_get_graph function computes a graph in networkx that represents these dependencies.
"""


def _get_varname(op):
    op_name = op.name
    idx = op_name.find('/')  # Use the first part of the operation name (until slash) as name
    if idx != -1 and '/Assign' not in op_name:  # Special case for tf.Variables
        return op_name[:idx]
    else:
        return op_name


def _children(op):
    # get the consumers of the operation as its children (set of names, using _get_varname)
    return set(_get_varname(opc) for out in op.outputs for opc in out.consumers())


def _clean_graph(G, varnames):
    # G is a networkx graph. Clean nodes from it whose names are not in varnames set or dict.
    # Before removing such nodes, create an edge between their parents and their children.
    g_nodes = list(G.nodes)
    for n in g_nodes:
        if n not in varnames:
            if '/Assign' in n:  # Special case for tf.Variables
                predecesors = list(G.predecessors(n))
                n_name = n[n.rfind('/')]  # real name of the tf.variable

                assert len(predecesors) <= 2  # At most, it should have two predecessors

                if len(predecesors) == 2:
                    if predecesors[0] == n_name:
                        # create relation from predecesors[1] to predecesors[0]
                        G.add_edge(predecesors[1], predecesors[0])
                    else:
                        # create relation from predecesors[0] to predecesors[1]
                        G.add_edge(predecesors[0], predecesors[1])
            else:
                # remove and create edge between parent and child if exist
                for p in G.predecessors(n):
                    for s in G.successors(n):
                        G.add_edge(p, s)
            G.remove_node(n)
    return G


def get_graph(varnames):
    # varnames is a set or dict where keys are the var names of the Random Variables
    if not (isinstance(varnames, dict) or isinstance(varnames, set)):
        raise TypeError("The type of varnames must be dict or set, not {}".format(type(varnames)))

    # Creates dictionary {node: {child1, child2, ..},..} for current
    # TensorFlow graph. Result is compatible with networkx/toposort
    # Uses the default_graph
    ops = tf.get_default_graph().get_operations()
    dependencies = defaultdict(set)
    for op in ops:
        # in tensorflow_probability, the tensor named *sample_shape* is a op in child-parent order.
        # as we want to capture only the parent-child relations, skip these op names
        if 'sample_shape' not in op.name:
            c = _children(op)
            op_name = _get_varname(op)
            c.discard(op_name)  # avoid name references to itself
            dependencies[op_name].update(c)

    # create networkx graph
    G = nx.DiGraph(dependencies)
    # clean names, to get just ed2 RV var names
    _clean_graph(G, varnames)  # inplace modification
    return G


def get_empty_graph():
    return nx.DiGraph()



# -*- coding: utf-8 -*-
#

"""
Package with modules defining functions, classes and variables which are
useful for the main functionality provided by inferpy
"""


from .runtime import tf_run_allowed, tf_run_ignored, set_tf_run
from . import iterables
from . import interceptor
from . import name
from .session import get_session, set_session, clear_session

__all__ = [
    'iterables',
    'interceptor',
    'set_tf_run',
    'tf_run_allowed',
    'tf_run_ignored',
    'name',
    'get_session',
    'set_session',
    'clear_session'
]
